def load_text(*, path_text: str):
    with open(path_text, "r", encoding="utf-8") as file:
        text = file.read()
    return text


class PunctuationChars:
    def __init__(self):
        self.allowed = set(" ,-.:;“”'‘’()!?…" + '"')


def exclude_invalid_chars(*, text: str):
    text = text.replace("\n", " ")
    text = text.replace("—", " ")
    text = text.replace("_", " ")
    text = text.replace(b"\xe3\x80\x80".decode("utf8"), " ")

    punctuation_chars = PunctuationChars().allowed

    alphanumeric_chars_allowed = set(
        "0123456789"
        + "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        + "abcdefghijklmnopqrstuvwxyz"
        + "ÁÂÃÄÅÆÇÈÉÊËÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝ"
        + "ÞßàáâãäåæçèéêëíîïðñòóôõöøùúûüýþÿŌōŞşŪūŸŻż"
    )

    chars_allowed = alphanumeric_chars_allowed | punctuation_chars

    text_filtered = []

    for idx_char in range(len(text)):
        char = text[idx_char]
        if char not in chars_allowed:
            # Remove invalid character.
            try:
                context = text[(idx_char - 10) : (idx_char + 10)]
            except KeyError:
                context = None
            print(f"Removing char: {char} | Context: {context}")
        else:
            # Add valid character to filtered text.
            text_filtered.append(char)

    text_filtered = "".join(text_filtered)

    return text_filtered


def load_and_preprocess_text(*, path_text: str):
    text = load_text(path_text=path_text)

    text = exclude_invalid_chars(text=text)

    # One large string to list of words.
    text = text.split(" ")

    # Remove empty strings.
    text = [x for x in text if len(x) > 0]

    return text



import random
from copy import copy


def words_identical(word_1: str, word_2: str):
    """
    Compare words, ignoring capitalization and punctuation. For example "Horse" and
    "horse" are counted as identical. Similarly, "horse," and "horse" are considered
    identical. The reason for this is that in the listening condition it would
    otherwise be difficult to identify target events (because "horse," and "horse" sound
    similar).
    """
    word_1 = word_1.lower()
    word_2 = word_2.lower()

    word_1 = "".join(char for char in word_1 if char.isalnum())
    word_2 = "".join(char for char in word_2 if char.isalnum())

    if word_1 == word_2:
        return True
    else:
        return False


def get_target_events(*, text: list[str]):
    """
    Check for target events (i.e. repeated words) in text.
    """
    target_event_idcs = []

    for idx_word in range(0, (len(text) - 1)):
        this_word = text[idx_word]
        next_word = text[idx_word + 1]
        if words_identical(this_word, next_word):
            target_event_idcs.append(idx_word + 1)

    return target_event_idcs


def check_targets_too_close(*, target_idcs: list[int], min_distance_targets: int):
    """
    Ensure that two target events do not directly follow each other.
    """
    targets_too_close = None

    target_idcs = sorted(target_idcs)

    # Loop over target events.
    for idx_sample in range(0, (len(target_idcs) - 1)):
        # Current target event (word index, e.g. 7 would correspond to the 7th word in
        # the text.
        target_idx = target_idcs[idx_sample]
        # Index of the subsequent target event.
        next_target_idx = target_idcs[idx_sample + 1]
        target_events_distance = next_target_idx - target_idx
        if target_events_distance < min_distance_targets:
            targets_too_close = target_idx
            break

    # Returns `None` or the integer index of a target (e.g. integer 7 if the 7th word in
    # the text is a target event) if that target event is followed by another target
    # event.
    return targets_too_close


def remove_double_repeats(*, text: list[str]):
    """
    Remove double repeated words (keep max 2 consecutive duplicates). For example, the
    input ["Apple", "Banana", "Banana", "Banana", "Coconut"] will get changed to
    ["Apple", "Banana", "Banana", "Coconut"]. In contrast, the input ["Apple", "Banana",
    "Banana", "Coconut"] will remained unchanged.
    """
    result = []

    for word in text:
        # Check if we already have at least 2 words in the result and if the last two
        # words match the current word.
        if (
            len(result) >= 2
            and words_identical(result[-1], word)
            and words_identical(result[-2], word)
        ):
            continue  # Skip adding this word

        result.append(word)

    return result


def sample_target_events(
    *,
    text: list[str],
    n_target_events: int,
    min_distance_targets: int = 3,
):
    n_words = len(text)
    if n_words <= (n_target_events * min_distance_targets * 2):
        raise AssertionError("Too many target events")

    # ----------------------------------------------------------------------------------
    # *** Process "naturally" occurring target events

    # Check for "naturally" occurring target events (i.e. repeated words) in the
    # original text.
    natural_target_event_idcs = get_target_events(text=text)
    n_natural_target_events = len(natural_target_event_idcs)
    print(
        f"Number of 'natural' target events (repeated words): {n_natural_target_events}"
    )

    # Remove potential double repeats (among 'natural' target events in the original
    # text).
    text = remove_double_repeats(text=text)

    # Check for "naturally" occurring target events (i.e. repeated words) in text after
    # removing double repeats.
    natural_target_event_idcs = get_target_events(text=text)
    n_natural_target_events = len(natural_target_event_idcs)
    print(
        "Number of 'natural' target events (repeated words) after removing potential "
        "double repeats: "
        f"{n_natural_target_events}"
    )

    # If necessary, remove double target events (note that previously, words that
    # repeated more than once were removed; e.g. "Apple Banana Banana Banana Coconut"
    # became "Apple Banana Banana Coconut"; but now, "Apple Banana Banana Coconut
    # Coconut Kiwi" becomes "Apple Banana Banana Coconut Kiwi".
    done = False
    while not done:
        natural_targets_too_close = check_targets_too_close(
            target_idcs=get_target_events(text=text),
            min_distance_targets=min_distance_targets,
        )

        if natural_targets_too_close is not None:
            context = " ".join(
                text[(natural_targets_too_close - 4) : (natural_targets_too_close + 4)]
            )
            print(
                "Natural target events too close, removing double target event: "
                f"{context}"
            )

            text.pop(natural_targets_too_close)  # Remove the xth word

            new_context = " ".join(
                text[(natural_targets_too_close - 4) : (natural_targets_too_close + 3)]
            )
            print(f"Context after removing double target event: {new_context}")

        else:
            done = True

    # Check for "naturally" occurring target events (i.e. repeated words) in text after
    # removing double target events.
    natural_target_event_idcs = get_target_events(text=text)
    n_natural_target_events = len(natural_target_event_idcs)
    print(
        "Number of 'natural' target events (repeated words) after removing potential "
        "double target events: "
        f"{n_natural_target_events}"
    )

    # ----------------------------------------------------------------------------------
    # *** Sample random target events

    # Update number of words after potentially removing natural / double target events.
    n_words = len(text)

    # Indices of target events (on a target event, the word will be repeated). For
    # example, a target event index of 7 means that the 7th word in the text will be a
    # target event (and will be repeated).

    done = False
    while not done:
        # Check if we need to add additional, randomly sampled target events (in
        # addition to potentially occurring "natural" target events in the original
        # text).
        if n_natural_target_events < n_target_events:
            # Inefficient solution, but doesn't matter as it only needs to run once,
            # offline.
            target_event_word_idcs = copy(natural_target_event_idcs)
            while len(target_event_word_idcs) < n_target_events:
                potential_sample = random.randrange(1, (n_words - 1))
                targets_too_close = check_targets_too_close(
                    target_idcs=copy(target_event_word_idcs) + [potential_sample],
                    min_distance_targets=min_distance_targets,
                )
                if targets_too_close:
                    continue
                else:
                    target_event_word_idcs.append(potential_sample)
        else:
            # There are enough 'natural' target events, no need to sample additional
            # ones.
            target_event_word_idcs = natural_target_event_idcs

        target_event_word_idcs = sorted(target_event_word_idcs)

        # Check whether the randomly sampled target events are too close to each other
        # or to the "natural" target events.
        targets_too_close = check_targets_too_close(
            target_idcs=target_event_word_idcs,
            min_distance_targets=min_distance_targets,
        )

        if targets_too_close:
            print("Target events too close. Will sample again.")
            continue
        else:
            # Targets are not too close, we can keep the randomly samples target event
            # indices.
            done = True
            break

    # ----------------------------------------------------------------------------------
    # *** Add the target events (repeated words) to the text

    # We need to add the randomly sampled target events (word repetitions) to the text,
    # but not the "naturally" occurring target events (because those are already in the
    # text).
    text_with_targets = []
    is_target = []  # Boolean list

    for idx_word, word in enumerate(text):
        if idx_word in natural_target_event_idcs:
            # The current word is a "natural" target event (i.e. a word repetition in
            # the original text).
            text_with_targets.append(word)
            is_target.append(True)
        elif idx_word in target_event_word_idcs:  # This has to be elif
            # This is a randomly sampled target event. Note that it is important that
            # this is not a "natural" target event, hence the `elif` after checking for
            # a "natural" target. Append the word twice (so that there is a repetition).
            text_with_targets.extend([word, word])
            # The second occurrence of the word is the target event.
            is_target.extend([False, True])
        else:
            # The current word is not a target event.
            text_with_targets.append(word)
            is_target.append(False)

    return {
        "text_with_targets": text_with_targets,
        "is_target": is_target,
    }



import time
import traceback

import pygame

pygame.mixer.init()

audio_files = [
    "/home/john/Dropbox/Deep_Learning/TTS/Qwen3-TTS-12Hz-1.7B-CustomVoice/daenerys_Sohee_None.ogg",
    "/home/john/Dropbox/Deep_Learning/TTS/Qwen3-TTS-12Hz-1.7B-CustomVoice/jon_Sohee_None.ogg",
    "/home/john/Dropbox/Deep_Learning/TTS/Qwen3-TTS-12Hz-1.7B-CustomVoice/nine_days_Sohee_None.ogg",
]
timestamps = []

rest_condition_color = [0, 0, 0]


running = True
while running:
    pygame.init()

    # Get screen dimensions and set up full screen.
    screen_info = pygame.display.Info()
    screen_width = screen_info.current_w
    screen_height = screen_info.current_h
    screen = pygame.display.set_mode((screen_width, screen_height), pygame.FULLSCREEN)
    pygame.display.set_caption("Silent Reading Experiment")
    pygame.mouse.set_visible(False)

    try:
        # Initial grey screen.
        pygame.time.wait(100)
        screen.fill(rest_condition_color)
        pygame.display.flip()
        pygame.time.wait(100)

        for filepath in audio_files:
            sound = pygame.mixer.Sound(filepath)
            start_time = time.time()
            channel = sound.play()

            # Wait until this sound finishes playing.
            while channel.get_busy():
                pygame.time.wait(10)  # avoid busy-waiting
            end_time = time.time()

            timestamps.append(
                {
                    "file": filepath,
                    "start": start_time,
                    "end": end_time,
                }
            )

        pygame.mixer.quit()

        for entry in timestamps:
            print(
                f"{entry['file']}: start={entry['start']:.3f}, end={entry['end']:.3f}"
            )

        running = False

    except Exception as e:
        print(f"An error occurred during the experiment: {e}")
        print(traceback.format_exc())
        running = False
    finally:
        pygame.quit()
        print("Experiment closed.")



import multiprocessing as mp
import os
import random
import traceback
from time import sleep

import numpy as np
import pygame

from nubrain.audio.tone import generate_tone
from nubrain.device.device_interface import create_eeg_device
from nubrain.experiment_text.data import eeg_data_logging
from nubrain.experiment_text.random_target_events import sample_target_events
from nubrain.experiment_text.text_config import TextConfig
from nubrain.misc.datetime import get_formatted_current_datetime
from nubrain.text.rendering import construct_fonts, render_spaced_text
from nubrain.text.tools import load_and_preprocess_text

mp.set_start_method("spawn", force=True)  # Necessary on if running on windows?


def experiment_text(config: dict):
    # ----------------------------------------------------------------------------------
    # *** Get config

    device_type = config["device_type"]
    lsl_stream_name = config.get("lsl_stream_name", "DSI-24")

    subject_id = config["subject_id"]
    session_id = config["session_id"]

    output_directory = config["output_directory"]
    path_text = config["path_text"]

    storage_bucket_name = config["storage_bucket_name"]
    storage_blob_name = config["storage_blob_name"]
    storage_bucket_credentials = config["storage_bucket_credentials"]

    eeg_channel_mapping = config.get("eeg_channel_mapping", None)

    utility_frequency = config["utility_frequency"]

    initial_rest_duration = config["initial_rest_duration"]
    stimulus_duration = config["stimulus_duration"]
    isi_duration = config["isi_duration"]
    isi_jitter = config["isi_jitter"]
    isi_extension_target = config["isi_extension_target"]
    inter_block_rest_duration = config["inter_block_rest_duration"]
    response_window_duration = config["response_window_duration"]

    word_idx_start = config["word_idx_start"]
    n_words_to_show = config["n_words_to_show"]
    n_target_events = config["n_target_events"]
    min_distance_targets = config["min_distance_targets"]
    stimuli_per_block = config["stimuli_per_block"]
    stimulus_font_sizes = config["stimulus_font_sizes"]
    stimulus_font_min_spacing = config["stimulus_font_min_spacing"]
    stimulus_font_max_spacing = config["stimulus_font_max_spacing"]

    eeg_device_address = config.get("eeg_device_address", None)

    text_config = TextConfig()

    # ----------------------------------------------------------------------------------
    # *** Test if output path exists

    if not os.path.isdir(output_directory):
        raise AssertionError(f"Target directory does not exist: {output_directory}")

    current_datetime = get_formatted_current_datetime()
    path_out_data = os.path.join(output_directory, f"eeg_{current_datetime}.h5")

    if os.path.isfile(path_out_data):
        raise AssertionError(f"Target file already exists: {path_out_data}")

    # ----------------------------------------------------------------------------------
    # *** Load text

    # Load text from file.
    text = load_and_preprocess_text(path_text=path_text)

    # Select subset of text.
    text = text[word_idx_start : (word_idx_start + n_words_to_show)]

    # Random target events. In case of a target event, the word will be repeated.
    text_and_targets = sample_target_events(
        text=text,
        n_target_events=n_target_events,
        min_distance_targets=min_distance_targets,
    )

    text = text_and_targets["text_with_targets"]
    is_target = text_and_targets["is_target"]

    # ----------------------------------------------------------------------------------
    # *** Prepare EEG measurement

    print(f"Initializing EEG device: {device_type}")

    device_kwargs = {"eeg_channel_mapping": eeg_channel_mapping}
    if device_type in ["cyton", "synthetic"]:
        device_kwargs["eeg_device_address"] = eeg_device_address
    elif device_type == "dsi24":
        device_kwargs["lsl_stream_name"] = lsl_stream_name
    else:
        raise ValueError(f"Unexpected `device_type`: {device_type}")

    eeg_device = create_eeg_device(device_type, **device_kwargs)

    eeg_device.prepare_session()

    # This is a bit clunky. At this point, `eeg_channel_mapping` is None or a dict with
    # a channel mapping from the config yaml file. Overwrite it with the channel mapping
    # from the device (in case of the DSI-24 device, the channel mapping from the device
    # is used in any case).
    eeg_channel_mapping = eeg_device.eeg_channel_mapping

    # Need to start the stream before calling `eeg_device.get_device_info()`, because
    # we retrieve data from board to determine data shape (number of channels).
    eeg_device.start_stream()
    sleep(0.1)

    # Get device info.
    device_info = eeg_device.get_device_info()
    eeg_board_description = device_info["board_description"]
    eeg_sampling_rate = device_info["sampling_rate"]
    eeg_channels = device_info["eeg_channels"]
    marker_channel = device_info["marker_channel"]
    n_channels_total = device_info["n_channels_total"]

    if device_type in ["cyton", "synthetic"]:
        # For Cyton device, we need to get the number of EEG channels from the device
        # (not sure, this might only work after starting the stream).
        eeg_device.eeg_channels = eeg_channels
        eeg_device.timestamp_channel = eeg_board_description["timestamp_channel"]

    print(f"Board: {eeg_board_description['name']}")
    print(f"Sampling Rate: {eeg_sampling_rate} Hz")
    print(f"EEG Channels: {eeg_channels}")
    print(f"Marker Channel: {marker_channel}")
    print(f"EEG Channel Mapping: {eeg_channel_mapping}")

    board_data, board_timestamps = eeg_device.get_board_data()

    print(f"Board data dtype: {board_data.dtype}")
    print(f"Board data shape: {board_data.shape}")
    print(f"Board timestamps shape: {board_timestamps.shape}")

    # ----------------------------------------------------------------------------------
    # *** Start data logging subprocess

    data_logging_queue = mp.Queue()

    subprocess_params = {
        "device_type": device_type,
        "subject_id": subject_id,
        "session_id": session_id,
        "path_text": path_text,
        # EEG parameters
        "eeg_board_description": eeg_board_description,
        "eeg_sampling_rate": eeg_sampling_rate,
        "n_channels_total": n_channels_total,
        "eeg_channels": eeg_channels,
        "marker_channel": marker_channel,
        "eeg_channel_mapping": eeg_channel_mapping,
        "eeg_device_address": eeg_device_address,
        # Timing parameters
        "initial_rest_duration": initial_rest_duration,
        "stimulus_duration": stimulus_duration,
        "isi_duration": isi_duration,
        "isi_jitter": isi_jitter,
        "isi_extension_target": isi_extension_target,
        "inter_block_rest_duration": inter_block_rest_duration,
        "response_window_duration": response_window_duration,
        # Experiment structure
        "word_idx_start": word_idx_start,
        "n_words_to_show": n_words_to_show,
        "n_target_events": n_target_events,
        "min_distance_targets": min_distance_targets,
        "stimuli_per_block": stimuli_per_block,
        "stimulus_font_sizes": stimulus_font_sizes,
        "stimulus_font_min_spacing": stimulus_font_min_spacing,
        "stimulus_font_max_spacing": stimulus_font_max_spacing,
        # Text and targets
        "text": text,  # List of str
        "is_target": is_target,  # List of bool
        # Storage
        "path_out_data": path_out_data,
        "storage_bucket_name": storage_bucket_name,
        "storage_blob_name": storage_blob_name,
        "storage_bucket_credentials": storage_bucket_credentials,
        # Misc
        "utility_frequency": utility_frequency,
        "data_logging_queue": data_logging_queue,
    }

    logging_process = mp.Process(target=eeg_data_logging, args=(subprocess_params,))
    logging_process.daemon = True
    logging_process.start()

    # ----------------------------------------------------------------------------------
    # *** Start experiment

    # Performance counters.
    n_hits = 0
    n_false_alarms = 0
    n_total_targets = sum(is_target)

    running = True
    while running:
        pygame.init()

        # ------------------------------------------------------------------------------
        # *** Prepare audio cues

        # Use an audio cue at the beginning and at the end of the inter-block interval,
        # so the participant can close their eyes / rest.

        # How long before the end of the inter-block interval to play the audio cue.
        pure_tone_end_delay = 1.0

        # Play the tone to cue the end of the inter-block interval x seconds before the
        # end of the inter-block interval.
        # Do not use the audio cue if the inter-block interval is too short.
        if inter_block_rest_duration <= (pure_tone_end_delay + 0.1):
            print(
                "WARNING: Will not use audio cue for the end of the inter-block "
                "interval because of short inter-block interval of "
                f"{inter_block_rest_duration} s"
            )
            use_ibi_audio_cue = False
        else:
            use_ibi_audio_cue = True

        if use_ibi_audio_cue:
            pygame.mixer.init(frequency=44100, size=-16, channels=2, buffer=512)

            # Get the sample rate from the mixer settings.
            sample_rate = pygame.mixer.get_init()[0]

            # Generate the tone data.
            tone_data_start = generate_tone(
                frequency=700,  # Pitch of the tone in Hz
                duration=0.3,  # Duration of audio cue
                amplitude=0.9,  # Volume, from 0.0 to 1.0
                sample_rate=sample_rate,
            )

            tone_data_end = generate_tone(
                frequency=1400,  # Pitch of the tone in Hz
                duration=0.3,  # Duration of audio cue
                amplitude=0.9,  # Volume, from 0.0 to 1.0
                sample_rate=sample_rate,
            )

            # Create a sound object from the numpy array.
            pure_tone_start = pygame.sndarray.make_sound(tone_data_start)
            pure_tone_end = pygame.sndarray.make_sound(tone_data_end)

        # ------------------------------------------------------------------------------
        # *** Prepare visual stimulus generation

        # Get screen dimensions and set up full screen.
        screen_info = pygame.display.Info()
        screen_width = screen_info.current_w
        screen_height = screen_info.current_h
        screen = pygame.display.set_mode(
            (screen_width, screen_height), pygame.FULLSCREEN
        )
        pygame.display.set_caption("Silent Reading Experiment")
        pygame.mouse.set_visible(False)

        try:
            # Initial grey screen.
            pygame.time.wait(100)
            screen.fill(text_config.rest_condition_color)
            pygame.display.flip()
            pygame.time.wait(100)
            screen.fill(text_config.rest_condition_color)
            pygame.display.flip()

            stimulus_fonts = construct_fonts(font_sizes=stimulus_font_sizes)

            # Clear board buffer.
            _, _ = eeg_device.get_board_data()

            # Pause for specified number of milliseconds.
            pygame.time.delay(int(round(initial_rest_duration * 1000.0)))

            # Send pre-stimulus EEG data (to avoid buffer overflow).
            eeg_data, eeg_ts = eeg_device.get_board_data()
            if eeg_data.size > 0:
                data_logging_queue.put(
                    {
                        "type": "eeg",
                        "eeg_data": eeg_data,
                        "eeg_timestamps": eeg_ts,
                    }
                )

            # Count stimuli to introduce breaks between blocks.
            stimulus_block_counter = 0

            # Loop through words.
            for word, is_target_event in zip(text, is_target):
                if not running:  # Check for quit event
                    break

                # Randomly sample a font (we render the stimulus using different fonts
                # to achieve different stimulus appearance in terms of low-level visual
                # features.
                font_data = random.choice(stimulus_fonts)
                font_name = font_data["font_name"]
                font_size = font_data["font_size"]
                font_is_bold = font_data["font_is_bold"]
                font_is_italic = font_data["font_is_italic"]
                font_color = random.choice(text_config.font_colors)
                font_spacing = np.random.uniform(
                    low=stimulus_font_min_spacing,
                    high=stimulus_font_max_spacing,
                )

                stimulus_text = render_spaced_text(
                    text=word,
                    font=font_data["font"],
                    color=font_color,
                    spacing=font_spacing,
                )

                stimulus_rect = stimulus_text.get_rect(
                    center=(screen_width // 2, screen_height // 2)
                )
                screen.blit(stimulus_text, stimulus_rect)
                pygame.display.flip()
                # Start of stimulus presentation.
                t_stim_start = eeg_device.lsl_local_clock()

                # When using an OpenBCI device, we insert a stimulus marker into the
                # time series data on the EEG board. These markers can be used during
                # analysis to identify stimulus events. For the DSI-24 device, we
                # instead use LSL timestamps stored in the hdf5 file for identifying
                # stimulus events.
                if device_type in ["cyton", "synthetic"]:
                    eeg_device.insert_marker(text_config.stim_start_marker)
                else:
                    data_logging_queue.put(
                        {
                            "type": "marker",
                            "marker_value": text_config.stim_start_marker,
                            "timestamp": t_stim_start,
                        }
                    )

                response_made = False
                response_time = np.nan
                response_deadline = t_stim_start + response_window_duration

                # Wait for stimulus duration, but check for responses continuously.
                t_stim_end_expected = t_stim_start + stimulus_duration
                while eeg_device.lsl_local_clock() < t_stim_end_expected:
                    for event in pygame.event.get():
                        if event.type == pygame.QUIT:
                            running = False
                        if event.type == pygame.KEYDOWN:
                            keydown_time = eeg_device.lsl_local_clock()
                            if event.key == pygame.K_ESCAPE:
                                running = False
                            # Check for space bar press within the response window.
                            if event.key == pygame.K_SPACE and not response_made:
                                if keydown_time < response_deadline:
                                    response_made = True
                                    response_time = keydown_time - t_stim_start
                                    print(f"Response time: {round(response_time, 3)}")
                                    if is_target_event:
                                        # Hit.
                                        n_hits += 1
                                    else:
                                        # False alarm.
                                        n_false_alarms += 1
                    if not running:
                        break
                if not running:
                    break

                # End of stimulus presentation. Display ISI grey screen.
                screen.fill(text_config.rest_condition_color)
                pygame.display.flip()
                t_stim_end_actual = eeg_device.lsl_local_clock()

                # When using an OpenBCI device, we insert a stimulus marker into the
                # time series data on the EEG board. These markers can be used during
                # analysis to identify stimulus events. For the DSI-24 device, we
                # instead use LSL timestamps stored in the hdf5 file for identifying
                # stimulus events.
                if device_type in ["cyton", "synthetic"]:
                    eeg_device.insert_marker(text_config.stim_end_marker)
                else:
                    data_logging_queue.put(
                        {
                            "type": "marker",
                            "marker_value": text_config.stim_end_marker,
                            "timestamp": t_stim_end_actual,
                        }
                    )

                eeg_data, eeg_ts = eeg_device.get_board_data()
                if eeg_data.size > 0:
                    data_logging_queue.put(
                        {
                            "type": "eeg",
                            "eeg_data": eeg_data,
                            "eeg_timestamps": eeg_ts,
                        }
                    )

                # Time until when to show grey screen (ISI).
                t_isi_end = (
                    t_stim_end_actual
                    + isi_duration
                    + np.random.uniform(low=0.0, high=isi_jitter)
                )

                if is_target_event:
                    # If this is a target event, prolong the ISI duration, to allow the
                    # subject to respond before the onset of the next stimulus, to
                    # reduce the probability of a motor response artefact in the
                    # subsequent trial.
                    t_isi_end += isi_extension_target

                # Continue checking for late responses or quit events.
                while eeg_device.lsl_local_clock() < t_isi_end:
                    for event in pygame.event.get():
                        if event.type == pygame.QUIT:
                            running = False
                        if event.type == pygame.KEYDOWN:
                            keydown_time = eeg_device.lsl_local_clock()
                            if event.key == pygame.K_ESCAPE:
                                running = False
                            # Still check for spacebar presses that are within the
                            # response window for target events.
                            if event.key == pygame.K_SPACE and not response_made:
                                if keydown_time < response_deadline:
                                    response_made = True
                                    response_time = keydown_time - t_stim_start
                                    print(f"Response time: {round(response_time, 3)}")
                                    if is_target_event:
                                        # Hit.
                                        n_hits += 1
                                    else:
                                        # False alarm.
                                        n_false_alarms += 1
                    if not running:
                        break
                if not running:
                    break

                stimulus_data = {
                    "stimulus_start_time": t_stim_start,
                    "stimulus_end_time": t_stim_end_actual,
                    "stimulus_duration_s": t_stim_end_actual - t_stim_start,
                    "word": word,
                    "font_name": font_name,
                    "font_size": font_size,
                    "font_is_bold": font_is_bold,
                    "font_is_italic": font_is_italic,
                    "font_spacing": font_spacing,
                    "font_color": font_color,
                    "is_target_event": is_target_event,
                    "response_time_s": response_time,
                }
                data_logging_queue.put(
                    {"type": "stimulus", "stimulus_data": stimulus_data}
                )

                if not running:
                    break

                # Send post-stimulus EEG data (to avoid buffer overflow).
                eeg_data, eeg_ts = eeg_device.get_board_data()
                if eeg_data.size > 0:
                    data_logging_queue.put(
                        {"type": "eeg", "eeg_data": eeg_data, "eeg_timestamps": eeg_ts}
                    )

                stimulus_block_counter += 1

                if stimulus_block_counter == stimuli_per_block:
                    # Inter-block grey screen.
                    screen.fill(text_config.rest_condition_color)
                    pygame.display.flip()
                    # Start of inter-block interval.
                    t_ibi_start = eeg_device.lsl_local_clock()

                    if use_ibi_audio_cue:
                        # Audio cue to signal the beginning of the inter-block interval.
                        pure_tone_start.play()

                    # End of inter-block interval.
                    t_ibi_end = t_ibi_start + inter_block_rest_duration

                    if use_ibi_audio_cue:
                        # Time when to play audio cue to signal end of inter-block
                        # interval.
                        t_ibi_end_audio_cue = t_ibi_end - pure_tone_end_delay
                        ibi_end_cue_played_yet = False

                    while (
                        eeg_device.lsl_local_clock() < t_ibi_end
                    ):  # Continue inter-block interval?
                        if (
                            use_ibi_audio_cue
                            and (t_ibi_end_audio_cue <= eeg_device.lsl_local_clock())
                            and not ibi_end_cue_played_yet
                        ):  # Time to play end of inter-block interval cue?
                            # Play the cue to signal the end of the inter-block
                            # interval.
                            pure_tone_end.play()
                            ibi_end_cue_played_yet = True

                    stimulus_block_counter = 0

                    # Send inter-block EEG data (to avoid buffer overflow).
                    eeg_data, eeg_ts = eeg_device.get_board_data()
                    if eeg_data.size > 0:
                        data_logging_queue.put(
                            {
                                "type": "eeg",
                                "eeg_data": eeg_data,
                                "eeg_timestamps": eeg_ts,
                            }
                        )

            # End of word loop.

            # Calculate behavioural results.
            n_misses = n_total_targets - n_hits

            # Write behavioural results to hdf5 file.
            behavioural_data = {
                "n_total_targets": n_total_targets,
                "n_hits": n_hits,
                "n_misses": n_misses,
                "n_false_alarms": n_false_alarms,
            }
            data_logging_queue.put(
                {"type": "behavioural", "behavioural_data": behavioural_data}
            )

            if running:
                # Display behavioural results.
                screen.fill(text_config.rest_condition_color)

                # Behavioural results title.
                title_font = pygame.font.Font(None, 72)
                title_text = title_font.render(
                    "Experiment Complete",
                    True,
                    text_config.font_colors[0],
                )
                title_rect = title_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 - 150)
                )
                screen.blit(title_text, title_rect)

                # Behavioural results text.
                results_font = pygame.font.Font(None, 56)
                hits_text = results_font.render(
                    f"Hits: {n_hits}",
                    True,
                    text_config.font_colors[0],
                )
                misses_text = results_font.render(
                    f"Misses: {n_misses}",
                    True,
                    text_config.font_colors[0],
                )
                false_alarms_text = results_font.render(
                    f"False Alarms: {n_false_alarms}",
                    True,
                    text_config.font_colors[0],
                )

                # Position and display results
                hits_rect = hits_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 - 20)
                )
                misses_rect = misses_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 + 40)
                )
                false_alarms_rect = false_alarms_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 + 100)
                )

                screen.blit(hits_text, hits_rect)
                screen.blit(misses_text, misses_rect)
                screen.blit(false_alarms_text, false_alarms_rect)

                pygame.display.flip()
                pygame.time.wait(5000)  # Show results for 5 seconds

            running = False

            # Send final board data.
            eeg_data, eeg_ts = eeg_device.get_board_data()
            if eeg_data.size > 0:
                data_logging_queue.put(
                    {"type": "eeg", "eeg_data": eeg_data, "eeg_timestamps": eeg_ts}
                )

        except Exception as e:
            print(f"An error occurred during the experiment: {e}")
            print(traceback.format_exc())
            running = False
        finally:
            pygame.quit()
            print("Experiment closed.")

    eeg_device.stop_stream()
    eeg_device.release_session()

    print("Join process for sending data")
    data_logging_queue.put(None)
    logging_process.join()



# Is this necessary?

# from . import load_config

# Wrap the main import in try, so that the other modules can be imported without
# dependency on pylsl (e.g. when using image_config during preprocessing).
# try:
#     from . import main
# except Exception as e:
#     print(f"Failed to import nubrain experiment_eeg_to_image_v1 main module: {e}")



"""
This version: Uses previous reconstructed image as new stimulus.
"""

import asyncio
import base64
import io
import json
import multiprocessing as mp
import os
import queue
import random
import threading
import traceback
from time import sleep, time

import numpy as np
import pygame
import websockets

from nubrain.audio.tone import generate_tone
from nubrain.device.device_interface import create_eeg_device
from nubrain.experiment_image.data import eeg_data_logging
from nubrain.experiment_image.image_config import ImageConfig
from nubrain.experiment_image.randomize_conditions import (
    sample_next_image,
)
from nubrain.image.tools import (
    get_all_images,
    load_and_scale_image,
    scale_image_surface,
)
from nubrain.misc.datetime import get_formatted_current_datetime

mp.set_start_method("spawn", force=True)  # Necessary on if running on windows?


def websocket_client_thread(uri, request_json, image_queue):
    """
    Handle WebSocket communication in a separate thread.

    Connects to the WebSocket, sends data, and puts received images into a queue.
    """

    async def client_logic():
        async with websockets.connect(uri, max_size=5242880) as websocket:
            # Send the EEG data
            await websocket.send(request_json)

            # Listen for incoming image messages
            while True:
                message_str = await websocket.recv()
                message = json.loads(message_str)

                if "error" in message:
                    print(f"Server error: {message['error']}")
                    image_queue.put(None)  # Signal error
                    break

                # Put the received data into the thread-safe queue
                image_queue.put(message)

                if message.get("step") == "final":
                    break  # End of stream

    # Run the async logic in a new event loop for this thread
    asyncio.run(client_logic())


def experiment_eeg_to_image_v1_autoregressive(config: dict):
    # ----------------------------------------------------------------------------------
    # *** Get config

    device_type = config["device_type"]
    lsl_stream_name = config.get("lsl_stream_name", "DSI-24")

    subject_id = config["subject_id"]
    session_id = config["session_id"]

    output_directory = config["output_directory"]
    image_directory = config["image_directory"]

    eeg_channel_mapping = config.get("eeg_channel_mapping", None)

    utility_frequency = config["utility_frequency"]

    initial_rest_duration = config["initial_rest_duration"]
    pre_stimulus_interval = config["pre_stimulus_interval"]
    image_duration = config["image_duration"]
    post_stimulus_interval = config["post_stimulus_interval"]
    generated_image_duration = config["generated_image_duration"]
    isi_jitter = config["isi_jitter"]
    inter_block_grey_duration = config["inter_block_grey_duration"]

    n_blocks = config["n_blocks"]
    images_per_block = config["images_per_block"]

    api_endpoint = config["api_endpoint"]

    output_dir_images = config["output_dir_images"]

    eeg_device_address = config.get("eeg_device_address", None)

    image_config = ImageConfig()

    # ----------------------------------------------------------------------------------
    # *** Test if output path exists

    if not os.path.isdir(output_directory):
        raise AssertionError(f"Target directory does not exist: {output_directory}")

    current_datetime = get_formatted_current_datetime()
    path_out_data = os.path.join(output_directory, f"eeg_{current_datetime}.h5")

    if os.path.isfile(path_out_data):
        raise AssertionError(f"Target file already exists: {path_out_data}")

    # ----------------------------------------------------------------------------------
    # *** Get input images & their categories

    images_and_categories = get_all_images(image_directory=image_directory)

    if not images_and_categories:
        raise AssertionError(f"Found no images at {image_directory}")
    print(f"Found {len(images_and_categories)} images")

    # ----------------------------------------------------------------------------------
    # *** Create pseudo-random condition order

    # List with all unique image categories (e.g. `["apple", "banana", ...]`).
    image_categories = list(set([x["image_category"] for x in images_and_categories]))

    # Pick one target category at random.
    source_image_category = random.choice(image_categories)

    # Mapping from image categories to image file paths, e.g. `{"apple":
    # ["/path/to/apple_1.png", "/path/to/apple_2.png", ...], "banana":
    # ["/path/to/banana_2.png", ...]}`.
    category_to_filepath = {}
    for item in images_and_categories:
        image_category = item["image_category"]
        image_filepath = item["image_file_path"]
        if image_category in category_to_filepath:
            category_to_filepath[image_category].append(image_filepath)
        else:
            category_to_filepath[image_category] = [image_filepath]

    previous_image_file_path = None

    # ----------------------------------------------------------------------------------
    # *** Prepare EEG measurement

    print(f"Initializing EEG device: {device_type}")

    device_kwargs = {"eeg_channel_mapping": eeg_channel_mapping}
    if device_type in ["cyton", "synthetic"]:
        device_kwargs["eeg_device_address"] = eeg_device_address
    elif device_type == "dsi24":
        device_kwargs["lsl_stream_name"] = lsl_stream_name
    else:
        raise ValueError(f"Unexpected `device_type`: {device_type}")

    eeg_device = create_eeg_device(device_type, **device_kwargs)

    eeg_device.prepare_session()

    # This is a bit clunky. At this point, `eeg_channel_mapping` is None or a dict with
    # a channel mapping from the config yaml file. Overwrite it with the channel mapping
    # from the device (in case of the DSI-24 device, the channel mapping from the device
    # is used in any case).
    eeg_channel_mapping = eeg_device.eeg_channel_mapping

    # Need to start the stream before calling `eeg_device.get_device_info()`, because
    # we retrieve data from board to determine data shape (number of channels).
    eeg_device.start_stream()
    sleep(0.1)

    # Get device info.
    device_info = eeg_device.get_device_info()
    eeg_board_description = device_info["board_description"]
    eeg_sampling_rate = device_info["sampling_rate"]
    eeg_channels = device_info["eeg_channels"]
    marker_channel = device_info["marker_channel"]
    n_channels_total = device_info["n_channels_total"]

    if device_type in ["cyton", "synthetic"]:
        # For Cyton device, we need to get the number of EEG channels from the device
        # (not sure, this might only work after starting the stream).
        eeg_device.eeg_channels = eeg_channels
        eeg_device.timestamp_channel = eeg_board_description["timestamp_channel"]

    print(f"Board: {eeg_board_description['name']}")
    print(f"Sampling Rate: {eeg_sampling_rate} Hz")
    print(f"EEG Channels: {eeg_channels}")
    print(f"Marker Channel: {marker_channel}")
    print(f"EEG Channel Mapping: {eeg_channel_mapping}")

    board_data, board_timestamps = eeg_device.get_board_data()

    print(f"Board data dtype: {board_data.dtype}")
    print(f"Board data shape: {board_data.shape}")
    print(f"Board timestamps shape: {board_timestamps.shape}")

    # ----------------------------------------------------------------------------------
    # *** Start data logging subprocess

    data_logging_queue = mp.Queue()

    subprocess_params = {
        "device_type": device_type,
        "subject_id": subject_id,
        "session_id": session_id,
        "image_directory": image_directory,
        # EEG parameters
        "eeg_board_description": eeg_board_description,
        "eeg_sampling_rate": eeg_sampling_rate,
        "n_channels_total": n_channels_total,
        "eeg_channels": eeg_channels,
        "marker_channel": marker_channel,
        "eeg_channel_mapping": eeg_channel_mapping,
        "eeg_device_address": eeg_device_address,
        # Timing parameters
        "initial_rest_duration": initial_rest_duration,
        "image_duration": image_duration,
        "isi_duration": 0.0,  # Dummy value. TODO: Adjust logging for live eeg-to-image
        "inter_block_grey_duration": inter_block_grey_duration,
        # Experiment structure
        "n_blocks": n_blocks,
        "images_per_block": images_per_block,
        # Misc
        "utility_frequency": utility_frequency,
        "path_out_data": path_out_data,
        "data_logging_queue": data_logging_queue,
    }

    logging_process = mp.Process(target=eeg_data_logging, args=(subprocess_params,))
    logging_process.daemon = True
    logging_process.start()

    # ----------------------------------------------------------------------------------
    # *** Get source image

    # Sample the next image.
    next_image_file_path = sample_next_image(
        next_image_category=source_image_category,
        category_to_filepath=category_to_filepath,
        previous_image_file_path=previous_image_file_path,
    )

    # ----------------------------------------------------------------------------------
    # *** Start experiment

    running = True
    while running:
        pygame.init()

        # ------------------------------------------------------------------------------
        # *** Prepare audio cue

        pygame.mixer.init(frequency=44100, size=-16, channels=2, buffer=512)

        frequency = 1000  # Pitch of the tone in Hz (e.g., 1000 Hz is a common beep)
        duration = 0.3  # Duration in seconds (half a second)
        amplitude = 0.5  # Volume, from 0.0 to 1.0

        # Play the tone x seconds before the end of the pre-stimulus period.
        tone_pre_stimulus_onset = 0.6

        # Get the sample rate from the mixer settings.
        sample_rate = pygame.mixer.get_init()[0]

        # Generate the tone data.
        tone_data = generate_tone(
            frequency=frequency,
            duration=duration,
            amplitude=amplitude,
            sample_rate=sample_rate,
        )

        # Create a sound object from the numpy array.
        pure_tone = pygame.sndarray.make_sound(tone_data)

        # ------------------------------------------------------------------------------
        # *** Prepare visual stimulus generation

        # Get screen dimensions and set up full screen.
        screen_info = pygame.display.Info()
        screen_width = screen_info.current_w
        screen_height = screen_info.current_h
        screen = pygame.display.set_mode(
            (screen_width, screen_height), pygame.FULLSCREEN
        )
        pygame.display.set_caption("Image Presentation Experiment")
        pygame.mouse.set_visible(False)

        # Prepare text.
        font = pygame.font.Font(None, 56)

        text_original = font.render("Original image", True, (0, 0, 0))
        text_reconstructed = font.render("Reconstructed from EEG", True, (0, 0, 0))

        text_original_rect = text_original.get_rect(
            center=((screen_width * 1 // 4), (screen_height // 4 - 50))
        )
        text_reconstructed_rect = text_reconstructed.get_rect(
            center=((screen_width * 3 // 4), (screen_height // 4 - 50))
        )

        screen.blit(text_original, text_original_rect)
        screen.blit(text_reconstructed, text_reconstructed_rect)

        try:
            # Initial grey screen.
            pygame.time.wait(100)
            screen.fill(image_config.rest_condition_color)
            pygame.display.flip()
            pygame.time.wait(100)
            screen.fill(image_config.rest_condition_color)
            pygame.display.flip()

            # Pause for specified number of milliseconds.
            pygame.time.delay(
                int(round((initial_rest_duration - tone_pre_stimulus_onset) * 1000.0))
            )

            # Clear board buffer.
            _, _ = eeg_device.get_board_data()

            # Block loop.
            for idx_block in range(n_blocks):
                # Average embedding vectors within blocks (across trials). Show x
                # repetitions of the same image, perform inference, and average the
                # embedding vector.

                # Load the next image.
                image_and_metadata = None
                while image_and_metadata is None:
                    image_and_metadata = load_and_scale_image(
                        image_file_path=next_image_file_path,
                        screen_width=screen_width,
                        screen_height=screen_height,
                    )

                current_image = image_and_metadata["image"]

                # Play tone to cue block start.
                pure_tone.play()
                pygame.time.delay(int(round(tone_pre_stimulus_onset * 1000.0)))

                eeg_data = []
                eeg_timestamps = []

                # `marker_data` is an array of shape [2, n_timepoints], where the [0, :]
                # corresponds to timestamps, and [1, :] corresponds to the marker values
                # (represented as a nested list here).
                marker_data = np.zeros((2, (images_per_block * 2)))

                # Image loop (within a block).
                for image_count in range(images_per_block):
                    if not running:  # Check for quit event
                        break

                    # ------------------------------------------------------------------
                    # *** Pre-stimulus interval

                    # Start of the pre-stimulus interval.
                    t_pre_stim_start = time()

                    img_rect = current_image.get_rect(
                        center=(screen_width // 2, screen_height // 2)
                    )
                    screen.fill(image_config.rest_condition_color)
                    screen.blit(current_image, img_rect)

                    # Wait until the end of the pre-stimulus period.
                    t_pre_stim_end = t_pre_stim_start + pre_stimulus_interval
                    while time() < t_pre_stim_end:
                        for event in pygame.event.get():
                            if event.type == pygame.QUIT:
                                running = False
                            if event.type == pygame.KEYDOWN:
                                if event.key == pygame.K_ESCAPE:
                                    running = False
                        if not running:
                            break
                    if not running:
                        break

                    # ------------------------------------------------------------------
                    # *** Stimulus period

                    pygame.display.flip()
                    t_stim_start = time()  # Start of stimulus presentation.

                    # Insert stimulus start marker and get its timestamp.
                    marker_val_stim_start, marker_ts_stim_start = (
                        eeg_device.insert_marker(image_config.stim_start_marker)
                    )
                    if marker_val_stim_start is not None:
                        data_logging_queue.put(
                            {
                                "type": "marker",
                                "marker_value": marker_val_stim_start,
                                "timestamp": marker_ts_stim_start,
                            }
                        )
                    marker_data[0, (image_count * 2)] = marker_ts_stim_start
                    marker_data[1, (image_count * 2)] = marker_val_stim_start

                    # Send pre-stimulus EEG data (to avoid buffer overflow).
                    eeg_data_pre_stim, eeg_ts_pre_stim = eeg_device.get_board_data()
                    if eeg_data_pre_stim.size > 0:
                        data_logging_queue.put(
                            {
                                "type": "eeg",
                                "eeg_data": eeg_data_pre_stim,
                                "eeg_timestamps": eeg_ts_pre_stim,
                            }
                        )
                    eeg_data.append(eeg_data_pre_stim)
                    eeg_timestamps.append(eeg_ts_pre_stim)

                    # Wait for image duration, but check for responses continuously.
                    t_stim_end_expected = t_stim_start + image_duration
                    while time() < t_stim_end_expected:
                        for event in pygame.event.get():
                            if event.type == pygame.QUIT:
                                running = False
                            if event.type == pygame.KEYDOWN:
                                if event.key == pygame.K_ESCAPE:
                                    running = False
                        if not running:
                            break
                    if not running:
                        break

                    # ------------------------------------------------------------------
                    # *** Post-stimulus period

                    # End of stimulus presentation. Display grey screen.
                    screen.fill(image_config.rest_condition_color)
                    pygame.display.flip()
                    t_stim_end_actual = time()

                    marker_val_stim_end, marker_ts_stim_end = eeg_device.insert_marker(
                        image_config.stim_end_marker
                    )
                    if marker_val_stim_end is not None:
                        data_logging_queue.put(
                            {
                                "type": "marker",
                                "marker_value": marker_val_stim_end,
                                "timestamp": marker_ts_stim_end,
                            }
                        )
                    marker_data[0, ((image_count * 2) + 1)] = marker_ts_stim_end
                    marker_data[1, ((image_count * 2) + 1)] = marker_val_stim_end

                    eeg_data_stim, eeg_ts_stim = eeg_device.get_board_data()
                    if eeg_data_stim.size > 0:
                        data_logging_queue.put(
                            {
                                "type": "eeg",
                                "eeg_data": eeg_data_stim,
                                "eeg_timestamps": eeg_ts_stim,
                            }
                        )
                    eeg_data.append(eeg_data_stim)
                    eeg_timestamps.append(eeg_ts_stim)

                    # Wait until the end of the post-stimulus period
                    t_post_stim_end = t_stim_end_actual + post_stimulus_interval
                    while time() < t_post_stim_end:
                        for event in pygame.event.get():
                            if event.type == pygame.QUIT:
                                running = False
                            if event.type == pygame.KEYDOWN:
                                if event.key == pygame.K_ESCAPE:
                                    running = False
                        if not running:
                            break
                    if not running:
                        break

                    eeg_data_post_stim, eeg_ts_post_stim = eeg_device.get_board_data()
                    if eeg_data_post_stim.size > 0:
                        data_logging_queue.put(
                            {
                                "type": "eeg",
                                "eeg_data": eeg_data_post_stim,
                                "eeg_timestamps": eeg_ts_post_stim,
                            }
                        )
                    eeg_data.append(eeg_data_post_stim)
                    eeg_timestamps.append(eeg_ts_post_stim)

                    # ------------------------------------------------------------------
                    # *** Log stimulus metadata

                    stimulus_data = {
                        "stimulus_start_time": t_stim_start,
                        "stimulus_end_time": t_stim_end_actual,
                        "stimulus_duration_s": t_stim_end_actual - t_stim_start,
                        "image_file_path": next_image_file_path,
                        "image_category": source_image_category,
                        "is_target_event": False,  # Dummy values, TODO: data logging for eeg-to-image
                        "response_time_s": 0.0,  # Dummy values, TODO: data logging for eeg-to-image
                    }
                    data_logging_queue.put(
                        {"type": "stimulus", "stimulus_data": stimulus_data}
                    )

                # ----------------------------------------------------------------------
                # *** Inference

                eeg_data = np.concatenate(eeg_data, axis=1)
                eeg_timestamps = np.concatenate(eeg_timestamps)

                request_dict = {
                    "eeg_data": eeg_data.tolist(),
                    "eeg_timestamps": eeg_timestamps.tolist(),
                    "marker_data": marker_data.tolist(),
                    "utility_frequency": utility_frequency,
                    "eeg_channel_mapping": eeg_channel_mapping,
                }

                request_json = json.dumps(request_dict)

                # Wueue to receive images from the websocket thread.
                image_queue = queue.Queue()

                client_thread = threading.Thread(
                    target=websocket_client_thread,
                    args=(api_endpoint, request_json, image_queue),
                    daemon=True,
                )
                client_thread.start()

                # ----------------------------------------------------------------------
                # *** Show generated images as they arrive next to the original image

                generated_image_surface = None
                path_image_out = None
                eeg_model_id = "unknown"
                is_first_message = True

                # Loop to display images as they are received from the thread.
                while True:
                    try:
                        # Check the queue for a new message (non-blocking).
                        message = image_queue.get_nowait()

                        if message is None:  # Error signal
                            print("Error receiving image from server.")
                            running = False
                            break

                        if is_first_message:
                            # The first message contains the model ID.
                            eeg_model_id = message.get("eeg_model_id", "unknown")
                            is_first_message = False
                            continue  # Wait for the next message which will be an image

                        # Decode the base64 image.
                        image_bytes = base64.b64decode(message["image_base64"])
                        image_file = io.BytesIO(image_bytes)

                        # Create a Pygame surface
                        generated_image_surface = pygame.image.load(
                            image_file
                        ).convert()

                        # Scale the image for display.
                        scaled_image_surface = scale_image_surface(
                            image_surface=generated_image_surface,  # TODO
                            screen_width=screen_width,
                            screen_height=screen_height,
                        )

                        # Display the original image on the left, and the generated
                        # image on the right.
                        original_img_rect = current_image.get_rect(
                            center=(
                                (screen_width * 1 // 4),
                                screen_height // 2 + 50,
                            )
                        )
                        generated_img_rect = scaled_image_surface.get_rect(
                            center=(
                                (screen_width * 3 // 4),
                                screen_height // 2 + 50,
                            )
                        )

                        screen.fill(image_config.rest_condition_color)
                        # Text titles (original & reconstructed image).
                        screen.blit(text_original, text_original_rect)
                        screen.blit(text_reconstructed, text_reconstructed_rect)
                        # Original image.
                        screen.blit(current_image, original_img_rect)
                        # Reconstructed image.
                        screen.blit(scaled_image_surface, generated_img_rect)

                        pygame.display.flip()

                        # If this is the final, high-quality image, save it.
                        if message.get("step") == "final":
                            time_now = get_formatted_current_datetime()
                            path_image_out = os.path.join(
                                output_dir_images,
                                f"{eeg_model_id}_{time_now}_{source_image_category}.png",
                            )
                            with open(path_image_out, "wb") as f:
                                f.write(image_bytes)
                            break  # Exit the image receiving loop

                    except queue.Empty:
                        # No new image in the queue, just continue the loop
                        pass

                    # Keep Pygame responsive
                    for event in pygame.event.get():
                        if event.type == pygame.QUIT or (
                            event.type == pygame.KEYDOWN
                            and event.key == pygame.K_ESCAPE
                        ):
                            running = False
                            break
                    if not running:
                        break

                if not running:
                    break

                # ----------------------------------------------------------------------
                # *** Show the final generated image for a fixed duration

                # Show generated image for this amount of time.
                t_generated_img_end = time() + generated_image_duration

                screen.fill(image_config.rest_condition_color)
                # Text titles (original & reconstructed image).
                screen.blit(text_original, text_original_rect)
                screen.blit(text_reconstructed, text_reconstructed_rect)
                # Original image.
                screen.blit(current_image, original_img_rect)
                # Reconstructed image.
                screen.blit(scaled_image_surface, generated_img_rect)

                pygame.display.flip()

                # Insert stimulus start marker and get its timestamp.
                generated_img_start_marker = 3.0  # Hardcoded TODO make config param
                marker_val_stim_start, marker_ts_stim_start = eeg_device.insert_marker(
                    generated_img_start_marker
                )
                if marker_val_stim_start is not None:
                    data_logging_queue.put(
                        {
                            "type": "marker",
                            "marker_value": marker_val_stim_start,
                            "timestamp": marker_ts_stim_start,
                        }
                    )

                while time() < t_generated_img_end:
                    for event in pygame.event.get():
                        if event.type == pygame.QUIT:
                            running = False
                        if event.type == pygame.KEYDOWN:
                            if event.key == pygame.K_ESCAPE:
                                running = False
                    if not running:
                        break
                if not running:
                    break

                # ----------------------------------------------------------------------
                # *** Log EEG data (to avoid buffer overflow)

                # Log data from the interval when waiting for inference results.

                eeg_data_wait, eeg_ts_wait = eeg_device.get_board_data()
                if eeg_data_wait.size > 0:
                    data_logging_queue.put(
                        {
                            "type": "eeg",
                            "eeg_data": eeg_data_wait,
                            "eeg_timestamps": eeg_ts_wait,
                        }
                    )

                # ----------------------------------------------------------------------
                # *** Grey screen (after generated image)

                # End of generated image presentation. Display grey screen.
                screen.fill(image_config.rest_condition_color)
                pygame.display.flip()

                # Insert stimulus start marker and get its timestamp.
                generated_img_end_marker = 4.0  # Hardcoded TODO make config param
                marker_val_stim_end, marker_ts_stim_end = eeg_device.insert_marker(
                    generated_img_end_marker
                )
                if marker_val_stim_end is not None:
                    data_logging_queue.put(
                        {
                            "type": "marker",
                            "marker_value": marker_val_stim_end,
                            "timestamp": marker_ts_stim_end,
                        }
                    )

                # ----------------------------------------------------------------------
                # *** Log EEG data (to avoid buffer overflow)

                eeg_data_gen_img, eeg_ts_gen_img = eeg_device.get_board_data()
                if eeg_data_gen_img.size > 0:
                    data_logging_queue.put(
                        {
                            "type": "eeg",
                            "eeg_data": eeg_data_gen_img,
                            "eeg_timestamps": eeg_ts_gen_img,
                        }
                    )

                if not running:
                    break

                # ----------------------------------------------------------------------
                # *** Prepare next block

                # Show the generated image as the new source image.
                next_image_file_path = path_image_out

                # Inter-block grey screen.
                screen.fill(image_config.rest_condition_color)
                pygame.display.flip()
                remaining_wait = (
                    inter_block_grey_duration
                    - tone_pre_stimulus_onset
                    + np.random.uniform(low=0.0, high=isi_jitter)
                )
                pygame.time.delay(int(round(remaining_wait * 1000.0)))

                # Send post-stimulus EEG data (to avoid buffer overflow).
                eeg_data, eeg_ts = eeg_device.get_board_data()
                if eeg_data.size > 0:
                    data_logging_queue.put(
                        {"type": "eeg", "eeg_data": eeg_data, "eeg_timestamps": eeg_ts}
                    )

            # --------------------------------------------------------------------------
            # *** End of experiment

            running = False

            # Send final board data.
            eeg_data, eeg_ts = eeg_device.get_board_data()
            if eeg_data.size > 0:
                data_logging_queue.put(
                    {"type": "eeg", "eeg_data": eeg_data, "eeg_timestamps": eeg_ts}
                )

        except Exception as e:
            print(f"An error occurred during the experiment: {e}")
            print(traceback.format_exc())
            running = False
        finally:
            pygame.quit()
            print("Experiment closed.")

    eeg_device.stop_stream()
    eeg_device.release_session()

    print("Join process for sending data")
    data_logging_queue.put(None)
    logging_process.join()



import json
import os
from time import time

import h5py
import numpy as np

from nubrain.experiment_text.text_config import TextConfig
from nubrain.storage.gcloud_bucket_upload import upload_to_gcs

text_config = TextConfig()


def eeg_data_logging(subprocess_params: dict):
    """
    Log experimental data.

    Continuously save to local hdf file. Upload to google cloud storage bucket at the
    end of the run. To be run in separate process (using multiprocessing).

    Please note that the stimulus onset and offset timestamps (`stimulus_start_time` and
    `stimulus_end_time`) use the LSL local clock, which is in seconds, but not aligned
    with UNIX epoch. We use the LSL clock for consistency with EEG timestamps from the
    DSI-24 device. Only the `experiment_start_time` is in UNIX epoch.
    """
    # ----------------------------------------------------------------------------------
    # *** Get parameters

    device_type = subprocess_params["device_type"]

    path_text = subprocess_params["path_text"]

    subject_id = subprocess_params["subject_id"]
    session_id = subprocess_params["session_id"]

    # EEG parameters
    eeg_board_description = subprocess_params["eeg_board_description"]
    eeg_sampling_rate = subprocess_params["eeg_sampling_rate"]
    n_channels_total = subprocess_params["n_channels_total"]
    eeg_channels = subprocess_params["eeg_channels"]
    marker_channel = subprocess_params["marker_channel"]
    eeg_channel_mapping = subprocess_params["eeg_channel_mapping"]
    eeg_device_address = subprocess_params["eeg_device_address"]

    # Timing parameters
    initial_rest_duration = subprocess_params["initial_rest_duration"]
    stimulus_duration = subprocess_params["stimulus_duration"]
    isi_duration = subprocess_params["isi_duration"]
    isi_jitter = subprocess_params["isi_jitter"]
    isi_extension_target = subprocess_params["isi_extension_target"]
    inter_block_rest_duration = subprocess_params["inter_block_rest_duration"]

    # Experiment structure
    word_idx_start = subprocess_params["word_idx_start"]
    n_words_to_show = subprocess_params["n_words_to_show"]
    n_target_events = subprocess_params["n_target_events"]
    min_distance_targets = subprocess_params["min_distance_targets"]
    stimuli_per_block = subprocess_params["stimuli_per_block"]
    stimulus_font_sizes = subprocess_params["stimulus_font_sizes"]
    stimulus_font_min_spacing = subprocess_params["stimulus_font_min_spacing"]
    stimulus_font_max_spacing = subprocess_params["stimulus_font_max_spacing"]

    # Text and targets
    text = subprocess_params["text"]  # List of str
    is_target = subprocess_params["is_target"]  # List of bool

    # Storage
    path_out_data = subprocess_params["path_out_data"]
    storage_bucket_name = subprocess_params["storage_bucket_name"]
    storage_blob_name = subprocess_params["storage_blob_name"]
    storage_bucket_credentials = subprocess_params["storage_bucket_credentials"]

    # Misc
    utility_frequency = subprocess_params["utility_frequency"]
    data_logging_queue = subprocess_params["data_logging_queue"]

    # ----------------------------------------------------------------------------------
    # *** Create and initialize HDF5 file

    experiment_metadata = {
        "config_version": text_config.config_version,
        "device_type": device_type,
        "subject_id": subject_id,
        "session_id": session_id,
        "path_text": path_text,
        "rest_condition_color": text_config.rest_condition_color,
        "stim_start_marker": text_config.stim_start_marker,
        "stim_end_marker": text_config.stim_end_marker,
        "hdf5_dtype": text_config.hdf5_dtype,
        "experiment_start_time": time(),  # Epoch timestamp
        # EEG parameters
        "eeg_board_description": eeg_board_description,
        "eeg_sampling_rate": eeg_sampling_rate,
        "n_channels_total": n_channels_total,
        "eeg_channel_mapping": eeg_channel_mapping,
        # Timing parameters
        "initial_rest_duration": initial_rest_duration,
        "stimulus_duration": stimulus_duration,
        "isi_duration": isi_duration,
        "isi_jitter": isi_jitter,
        "isi_extension_target": isi_extension_target,
        "inter_block_rest_duration": inter_block_rest_duration,
        # Experiment structure
        "word_idx_start": word_idx_start,
        "n_words_to_show": n_words_to_show,
        "n_target_events": n_target_events,
        "min_distance_targets": min_distance_targets,
        "stimuli_per_block": stimuli_per_block,
        "stimulus_font_sizes": stimulus_font_sizes,
        "stimulus_font_min_spacing": stimulus_font_min_spacing,
        "stimulus_font_max_spacing": stimulus_font_max_spacing,
        # Text and targets
        "text": text,  # List of str
        "is_target": is_target,  # List of bool
        # Misc
        "utility_frequency": utility_frequency,
    }

    # Parameters not used by DSI-24, for compatibility with Cyton board.
    if eeg_device_address is not None:
        experiment_metadata["eeg_device_address"] = eeg_device_address
    if eeg_channels is not None:
        experiment_metadata["eeg_channels"] = eeg_channels
    if marker_channel is not None:
        experiment_metadata["marker_channel"] = marker_channel

    print(f"Initializing HDF5 file at: {path_out_data}")
    with h5py.File(path_out_data, "w") as file:
        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for metadata

        # Create group for metadata.
        metadata_group = file.create_group("metadata")

        # Iterate over the Python dictionary and save each item as an attribute of the
        # "metadata" group.
        for key, value in experiment_metadata.items():
            # HDF5 attributes have limitations on data types. Complex types like
            # dictionaries or tuples are not natively supported. We check if the value
            # is a type that needs to be converted to a string. JSON is a convenient
            # format for this serialization.
            if isinstance(value, (dict, list, tuple)):
                # Serialize the complex type into a JSON string.
                metadata_group.attrs[key] = json.dumps(value)
            else:
                metadata_group.attrs[key] = value

        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for EEG data

        # Initialize dataset for EEG and additional channels. To handle a variable
        # number of timesteps, create a resizable dataset. We specify an initial shape
        # but set the 'maxshape' to allow one of the dimensions to be unlimited (by
        # setting it to None). 'chunks=True' is recommended for resizable datasets for
        # better performance. It lets h5py decide the chunk size.

        file.create_dataset(
            "eeg_data",
            shape=(n_channels_total, 0),
            maxshape=(n_channels_total, None),  # fixed_channels, unlimited_timesteps
            dtype=text_config.hdf5_dtype,
            chunks=True,
        )

        file.create_dataset(
            "eeg_timestamps",
            shape=(0,),
            maxshape=(None,),
            dtype="float64",
            chunks=True,
        )

        file.create_dataset(
            "marker_data",
            shape=(2, 0),  # timestamp, marker value
            maxshape=(2, None),
            dtype="float64",
            chunks=True,
        )

        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for stimulus data

        # Define the compound datatype for stimulus data. This is like defining the
        # columns of a table.
        stimulus_dtype = np.dtype(
            [
                ("stimulus_start_time", np.float64),
                ("stimulus_end_time", np.float64),
                ("stimulus_duration_s", np.float64),
                ("word", h5py.string_dtype(encoding="utf-8")),
                ("font_name", h5py.string_dtype(encoding="utf-8")),
                ("font_size", np.int64),
                ("font_is_bold", np.bool),
                ("font_is_italic", np.bool),
                ("font_color_r", np.uint8),
                ("font_color_g", np.uint8),
                ("font_color_b", np.uint8),
                ("font_spacing", np.float64),
                ("is_target_event", np.bool),
                ("response_time_s", np.float64),
            ]
        )

        file.create_dataset(
            "stimulus_data",
            (len(text),),
            dtype=stimulus_dtype,
        )

        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for behavioural data

        behavioural_dtype = np.dtype(
            [
                ("n_total_targets", np.int64),
                ("n_hits", np.int64),
                ("n_misses", np.int64),
                ("n_false_alarms", np.int64),
            ]
        )

        file.create_dataset(
            "behavioural_data",
            (1,),
            dtype=behavioural_dtype,
        )

    # ----------------------------------------------------------------------------------
    # *** Experiment loop

    stimulus_counter = 0

    while True:
        new_data = data_logging_queue.get(block=True)

        if new_data is None:
            # Received None. End process.
            print("Ending preprocessing & data saving process.")
            break

        data_type = new_data["type"]

        with h5py.File(path_out_data, "a") as file:
            # --------------------------------------------------------------------------
            # *** Write EEG data to hdf5 file

            if data_type == "eeg":
                new_eeg_data = new_data.get("eeg_data")
                new_timestamps = new_data.get("eeg_timestamps")

                if new_eeg_data is not None and new_eeg_data.size > 0:
                    # Write EEG data.
                    hdf5_eeg_data = file["eeg_data"]
                    n_existing = hdf5_eeg_data.shape[1]
                    n_new = new_eeg_data.shape[1]
                    hdf5_eeg_data.resize(n_existing + n_new, axis=1)
                    hdf5_eeg_data[:, n_existing:] = new_eeg_data

                    # Write EEG timestamps.
                    hdf5_timestamps = file["eeg_timestamps"]
                    n_existing_ts = hdf5_timestamps.shape[0]
                    hdf5_timestamps.resize(n_existing_ts + n_new, axis=0)
                    hdf5_timestamps[n_existing_ts:] = new_timestamps

            # --------------------------------------------------------------------------
            # *** Write stimulus markers to hdf5 file

            elif data_type == "marker":
                marker_value = new_data.get("marker_value")
                marker_timestamp = new_data.get("timestamp")

                if marker_value is not None:
                    hdf5_marker_data = file["marker_data"]
                    n_existing = hdf5_marker_data.shape[1]
                    hdf5_marker_data.resize(n_existing + 1, axis=1)
                    hdf5_marker_data[:, n_existing] = (marker_timestamp, marker_value)

            # --------------------------------------------------------------------------
            # *** Write stimulus data to hdf5 file

            elif data_type == "stimulus":
                new_stimulus_data = new_data.get("stimulus_data")

                if new_stimulus_data is not None:
                    hdf5_stimulus_data = file["stimulus_data"]

                    stimulus_start_time = new_stimulus_data["stimulus_start_time"]
                    stimulus_end_time = new_stimulus_data["stimulus_end_time"]
                    stimulus_duration_s = new_stimulus_data["stimulus_duration_s"]

                    word = new_stimulus_data["word"]
                    font_name = new_stimulus_data["font_name"]
                    font_size = new_stimulus_data["font_size"]
                    font_is_bold = new_stimulus_data["font_is_bold"]
                    font_is_italic = new_stimulus_data["font_is_italic"]
                    font_color = new_stimulus_data["font_color"]
                    font_spacing = new_stimulus_data["font_spacing"]

                    is_target_event = new_stimulus_data["is_target_event"]
                    response_time_s = new_stimulus_data["response_time_s"]

                    data_to_write = np.empty((1,), dtype=stimulus_dtype)

                    data_to_write[0]["stimulus_start_time"] = stimulus_start_time
                    data_to_write[0]["stimulus_end_time"] = stimulus_end_time
                    data_to_write[0]["stimulus_duration_s"] = stimulus_duration_s

                    data_to_write[0]["word"] = word
                    data_to_write[0]["font_name"] = font_name
                    data_to_write[0]["font_size"] = font_size
                    data_to_write[0]["font_is_bold"] = font_is_bold
                    data_to_write[0]["font_is_italic"] = font_is_italic
                    data_to_write[0]["font_color_r"] = font_color[0]
                    data_to_write[0]["font_color_g"] = font_color[1]
                    data_to_write[0]["font_color_b"] = font_color[2]

                    data_to_write[0]["font_spacing"] = font_spacing

                    data_to_write[0]["is_target_event"] = is_target_event
                    data_to_write[0]["response_time_s"] = response_time_s

                    # Write the structured array to the dataset.
                    hdf5_stimulus_data[stimulus_counter] = data_to_write

                    stimulus_counter += 1

            # --------------------------------------------------------------------------
            # *** Write behavioural data to hdf5 file

            elif data_type == "behavioural":
                new_behavioural_data = new_data.get("behavioural_data")

                if new_behavioural_data is not None:
                    hdf5_behavioural_data = file["behavioural_data"]

                    n_total_targets = new_behavioural_data["n_total_targets"]
                    n_hits = new_behavioural_data["n_hits"]
                    n_misses = new_behavioural_data["n_misses"]
                    n_false_alarms = new_behavioural_data["n_false_alarms"]

                    data_to_write = np.empty((1,), dtype=behavioural_dtype)

                    data_to_write[0]["n_total_targets"] = n_total_targets
                    data_to_write[0]["n_hits"] = n_hits
                    data_to_write[0]["n_misses"] = n_misses
                    data_to_write[0]["n_false_alarms"] = n_false_alarms

                    # Write the structured array to the dataset.
                    hdf5_behavioural_data[0] = data_to_write

    # ----------------------------------------------------------------------------------
    # *** Upload to cloud storage

    # Upload hdf5 file to google cloud storage bucket at the end of the run.

    filename = os.path.split(path_out_data)[-1]

    _storage_blob_name = storage_blob_name.format(
        device_type=device_type,
        filename=filename,
    )

    upload_to_gcs(
        local_file_path=path_out_data,
        bucket_name=storage_bucket_name,
        destination_blob_name=_storage_blob_name,
        credentials_file_path=storage_bucket_credentials,
    )



import pygame


def construct_fonts(
    *,
    font_sizes: list[int],
) -> list[dict]:
    """
    Generate a list of pygame fonts with different font types and settings (normal,
    bold, italic). Used to render text with different appearance in terms of low-level
    visual features.

    Note: You need to run `pygame.init()` in the parent function before constructing the
    fonts.
    """
    font_names = [
        "andalemono",
        "arial",
        "arialblack",
        "c059",
        "comicsansms",
        "couriernew",
        "dejavuserif",
        "georgia",
        "impact",
        "liberationmono",
        "liberationsans",
        "liberationsansnarrow",
        "liberationserif",
        "nimbusmonops",
        "nimbusroman",
        "nimbussans",
        "nimbussansnarrow",
        "notomono",
        "notosans",
        "notosanscjkhk",
        "notosansdisplay",
        "notosansmono",
        "notosansmonocjktc",
        "notoserif",
        "notoserifcjksc",
        "notoserifdisplay",
        "p052",
        "timesnewroman",
        "trebuchetms",
        "tuffy",
        "ubuntu",
        "ubuntumono",
        "ubuntusansmono",
        "urwbookman",
        "urwgothic",
        "verdana",
        "z003",
    ]

    fonts = []

    for font_name in font_names:
        for font_size in font_sizes:
            # Some bold fonts are not rendered nicely, let's skip bold.
            # for is_bold in [True, False]:
            is_bold = False
            for is_italic in [True, False]:
                new_font = pygame.font.SysFont(
                    font_name,
                    font_size,
                    is_bold,
                    is_italic,
                )
                fonts.append(
                    {
                        "font_name": font_name,
                        "font_size": font_size,
                        "font_is_bold": is_bold,
                        "font_is_italic": is_italic,
                        "font": new_font,
                    }
                )

    return fonts


def render_spaced_text(
    *,
    text: str,
    font: pygame.font.SysFont,
    color: tuple,
    spacing: float,
) -> pygame.Surface:
    """
    Renders text with custom letter spacing.

    Checks the width of every specific character as it is rendered. This is necessary
    because the width of characters differs between characters and fonts. Returns a
    pygame surface containing the spaced text.
    """
    if not text:
        return pygame.Surface((0, 0))

    # Render each character to its own surface.
    char_surfaces = [font.render(char, True, color) for char in text]

    # Calculate the total width and maximum height needed. Total width = sum of all char
    # widths + spacing between each char.
    total_width = sum([surf.get_width() for surf in char_surfaces]) + (
        spacing * (len(text) - 1)
    )
    max_height = max([surf.get_height() for surf in char_surfaces])

    # Create a master transparent surface to hold the word. SRCALPHA ensures the
    # background of this new surface remains transparent.
    word_surface = pygame.Surface((total_width, max_height), pygame.SRCALPHA)

    # Blit each character onto the master surface with the appropriate spacing offset.
    current_x = 0
    for surf in char_surfaces:
        word_surface.blit(surf, (current_x, 0))
        current_x += surf.get_width() + spacing

    return word_surface



import numpy as np


def generate_tone(
    *,
    frequency: int,
    duration: float,
    amplitude: float,
    sample_rate: int,
):
    """Generates a numpy array containing a sine wave for a pure tone."""
    # Calculate the number of samples needed.
    n_samples = int(round(sample_rate * duration))

    # Create an array representing time points.
    time_points = np.linspace(0, duration, n_samples, endpoint=False)

    # Generate the sine wave.
    # 2 * pi * frequency * t gives the angle for the sine function.
    wave = np.sin(2 * np.pi * frequency * time_points)

    # Scale wave to 16-bit integer range (-32767 to 32767) and apply amplitude. This
    # format is required by Pygame's mixer.
    wave = np.clip(
        np.around(wave * amplitude * 32767.0),
        a_min=-32767.0,
        a_max=32767.0,
    ).astype(np.int16)

    # Create a 2D array for stereo sound by duplicating the mono wave. Pygame's sndarray
    # requires a (num_samples, 2) array for stereo.
    stereo_wave = np.column_stack((wave, wave))

    return stereo_wave



from google.cloud import storage
from google.oauth2 import service_account


def upload_to_gcs(
    local_file_path: str,
    bucket_name: str,
    destination_blob_name: str,
    credentials_file_path: str,
) -> None:
    """
    Upload local file to google cloud storage bucket.

    Can be used to upload hdf5 file at the end of each run. Requires a service account
    key stored in a local JSON file.
    """
    try:
        # Authenticate using the JSON key file.
        credentials = service_account.Credentials.from_service_account_file(
            credentials_file_path
        )
        client = storage.Client(credentials=credentials)
        bucket = client.bucket(bucket_name)

        # Create a "blob" (the GCS equivalent of an S3 Object).
        blob = bucket.blob(destination_blob_name)

        print(f"Uploading: {local_file_path}")
        print(f"       to: gs://{bucket_name}/{destination_blob_name}")
        blob.upload_from_filename(local_file_path)

    except Exception as e:
        print(f"Error while uploading to google cloud storage: {e}")



import ctypes
import os
import sys
import tkinter as tk
from pathlib import Path
from tkinter import font, ttk

import yaml


def enable_dpi_awareness():
    """
    Tell Windows to report true pixel dimensions.

    If Windows desktop scaling is set to e.g. 150%, then SetProcessDpiAwareness(1) will
    report the true 4K pixel count rather than the scaled-down logical size. Without the
    DPI call, a 4K screen at 150% scaling would report itself as ~2560×1440, and the GUI
    would end up smaller than intended.
    """
    if sys.platform == "win32":
        try:
            ctypes.windll.shcore.SetProcessDpiAwareness(1)
        except Exception:
            try:
                ctypes.windll.user32.SetProcessDPIAware()
            except Exception:
                pass


class SessionConfigEditor:
    """GUI for editing experiment parameters (loaded from yaml file) before launch."""

    # Only these keys are shown in the GUI (and in this order)
    EDITABLE_FIELDS = [
        "subject",
        "session",
        "run",
    ]

    # Target fraction of screen real estate
    WIDTH_FRACTION = 0.3
    HEIGHT_FRACTION = 0.2

    def __init__(self, *, session_config_path: str):
        self.session_config_path = Path(session_config_path)
        with open(self.session_config_path, "rb") as file:
            self.session_config = yaml.safe_load(file)

        self.start_button_press = False

        enable_dpi_awareness()
        self.root = tk.Tk()
        self.root.title("NuBrain Experiment Configuration")
        self._scale_to_screen()
        self._build_gui()

    def _scale_to_screen(self):
        screen_w = self.root.winfo_screenwidth()
        screen_h = self.root.winfo_screenheight()

        win_w = int(screen_w * self.WIDTH_FRACTION)
        win_h = int(screen_h * self.HEIGHT_FRACTION)

        # Centre the window on screen.
        x = (screen_w - win_w) // 2
        y = (screen_h - win_h) // 2
        self.root.geometry(f"{win_w}x{win_h}+{x}+{y}")

        # Scale font size relative to screen height. ~14px looks good at 1080p
        # (height=1080), so use that as the anchor.
        base_font_size = max(10, int(screen_h / 1080 * 14))
        self.ui_font = font.Font(family="Segoe UI", size=base_font_size)
        self.heading_font = font.Font(
            family="Segoe UI", size=int(base_font_size * 1.3), weight="bold"
        )

        # Apply to all ttk widgets via a style.
        style = ttk.Style()
        style.configure("TLabel", font=self.ui_font)
        style.configure("TButton", font=self.ui_font)
        style.configure("TEntry", font=self.ui_font)

        # ttk.Entry doesn't always pick up the style font for the typed text, so set the
        # default font for Entry explicitly.
        self.root.option_add("*TEntry*Font", self.ui_font)

        # Store for later use in widget creation
        self.pad_x = max(10, int(screen_w * 0.008))
        self.pad_y = max(5, int(screen_h * 0.005))

    def _build_gui(self):
        self.entries = {}

        # Let column 1 (the entries) expand to fill available width.
        self.root.columnconfigure(1, weight=1)

        for i, key in enumerate(self.EDITABLE_FIELDS):
            ttk.Label(self.root, text=key).grid(
                row=i, column=0, padx=self.pad_x, pady=self.pad_y, sticky="e"
            )
            entry = ttk.Entry(self.root)
            entry.insert(0, str(self.session_config.get(key, "")))
            entry.grid(row=i, column=1, padx=self.pad_x, pady=self.pad_y, sticky="ew")
            self.entries[key] = entry

        btn_frame = ttk.Frame(self.root)
        btn_frame.grid(
            row=len(self.EDITABLE_FIELDS), column=0, columnspan=2, pady=self.pad_y * 3
        )
        ttk.Button(btn_frame, text="Start Experiment", command=self._on_start).pack(
            side="left", padx=self.pad_x
        )
        ttk.Button(btn_frame, text="Cancel", command=self.root.destroy).pack(
            side="left", padx=self.pad_x
        )

    def _on_start(self):
        self.start_button_press = True
        # Write GUI values back into the config dict.
        for key, entry in self.entries.items():
            value = entry.get()
            self.session_config[key] = value

        self.root.destroy()

    def input_value_validation(self):
        """All values (subject, session, run) have to be positive integers."""
        validated_session_config = {}
        for key, value in self.session_config.items():
            # Even if the user enters integer values, they might be represented as
            # strings at this point.
            try:
                value = int(value)
            except ValueError:
                pass  # Keep original value for error message
            if not isinstance(value, int):
                error_msg = (
                    f"Session config parameter '{key}' has to be an integer, "
                    f"got {type(value)}"
                )
                raise ValueError(error_msg)
            if value < 1:
                error_msg = (
                    f"Session config parameter '{key}' has to be an integer greater "
                    f"than zero, got {value}"
                )
                raise ValueError(error_msg)
            validated_session_config[key] = value
        self.session_config = validated_session_config

    def run(self) -> dict | None:
        """Show the GUI. Returns the edited config dict, or None if cancelled."""
        # Tkinter and pygame (in `main.py`) both want to own the main loop, so we have
        # to close the tkinter window (`self.root.destroy()` and letting `mainloop()`
        # return) before calling `pygame.init()` (in `main.py`). The below `mainloop()`
        # blocks until the window is destroyed, and only then does control flow to
        # `main.py`.
        self.root.mainloop()

        if self.start_button_press:
            # Validate user input.
            self.input_value_validation()

            # Write changes made in GUI back to file.
            with open(self.session_config_path, "w", encoding="utf-8") as file:
                yaml.safe_dump(self.session_config, file)

            return self.session_config

        else:
            # User pressed "Cancel".
            return None


if __name__ == "__main__":
    # DEBUGGING
    session_config_path = os.path.join(os.path.dirname(__file__), "session_config.yaml")

    # Show GUI
    gui = SessionConfigEditor(session_config_path=session_config_path)
    session_config = gui.run()
    if session_config is None:
        print("Cancelled.")
    else:
        print(f"Edited session_config: {session_config}")

    # Map from session config to experiment config.
    if session_config is not None:
        from map_config import map_session_config_to_experiment_config

        experiment_config = {"n_words_to_show": 600}
        updated_experiment_config = map_session_config_to_experiment_config(
            session_config=session_config,
            experiment_config=experiment_config,
        )
        print(f"updated_experiment_config: {updated_experiment_config}")



"""
Demo mode for instructing participants. Does not use EEG device, only presents a short
series of text stimuli (i.e. words).
"""

import random
import traceback
from time import time

import numpy as np
import pygame

from nubrain.audio.tone import generate_tone
from nubrain.experiment_text.random_target_events import sample_target_events
from nubrain.experiment_text.text_config import TextConfig
from nubrain.text.rendering import construct_fonts, render_spaced_text
from nubrain.text.tools import load_and_preprocess_text


def text_demo(config: dict):
    # ----------------------------------------------------------------------------------
    # *** Get config

    device_type = config["device_type"]

    path_text = config["path_text"]

    initial_rest_duration = config["initial_rest_duration"]
    stimulus_duration = config["stimulus_duration"]
    isi_duration = config["isi_duration"]
    isi_jitter = config["isi_jitter"]
    isi_extension_target = config["isi_extension_target"]
    inter_block_rest_duration = config["inter_block_rest_duration"]
    response_window_duration = config["response_window_duration"]

    word_idx_start = config["word_idx_start"]
    n_words_to_show = config["n_words_to_show"]
    n_target_events = config["n_target_events"]
    min_distance_targets = config["min_distance_targets"]
    stimuli_per_block = config["stimuli_per_block"]
    stimulus_font_sizes = config["stimulus_font_sizes"]
    stimulus_font_min_spacing = config["stimulus_font_min_spacing"]
    stimulus_font_max_spacing = config["stimulus_font_max_spacing"]

    text_config = TextConfig()

    # ----------------------------------------------------------------------------------
    # *** Load text

    # Load text from file.
    text = load_and_preprocess_text(path_text=path_text)

    # Select subset of text.
    text = text[word_idx_start : (word_idx_start + n_words_to_show)]

    # Random target events. In case of a target event, the word will be repeated.
    text_and_targets = sample_target_events(
        text=text,
        n_target_events=n_target_events,
        min_distance_targets=min_distance_targets,
    )

    text = text_and_targets["text_with_targets"]
    is_target = text_and_targets["is_target"]

    # ----------------------------------------------------------------------------------
    # *** Start experiment

    # Performance counters.
    n_hits = 0
    n_false_alarms = 0
    n_total_targets = sum(is_target)

    running = True
    while running:
        pygame.init()

        # ------------------------------------------------------------------------------
        # *** Prepare audio cues

        # Use an audio cue at the beginning and at the end of the inter-block interval,
        # so the participant can close their eyes / rest.

        # How long before the end of the inter-block interval to play the audio cue.
        pure_tone_end_delay = 1.0

        # Play the tone to cue the end of the inter-block interval x seconds before the
        # end of the inter-block interval.
        # Do not use the audio cue if the inter-block interval is too short.
        if inter_block_rest_duration <= (pure_tone_end_delay + 0.1):
            print(
                "WARNING: Will not use audio cue for the end of the inter-block "
                "interval because of short inter-block interval of "
                f"{inter_block_rest_duration} s"
            )
            use_ibi_audio_cue = False
        else:
            use_ibi_audio_cue = True

        if use_ibi_audio_cue:
            pygame.mixer.init(frequency=44100, size=-16, channels=2, buffer=512)

            # Get the sample rate from the mixer settings.
            sample_rate = pygame.mixer.get_init()[0]

            # Generate the tone data.
            tone_data_start = generate_tone(
                frequency=700,  # Pitch of the tone in Hz
                duration=0.3,  # Duration of audio cue
                amplitude=0.9,  # Volume, from 0.0 to 1.0
                sample_rate=sample_rate,
            )

            tone_data_end = generate_tone(
                frequency=1400,  # Pitch of the tone in Hz
                duration=0.3,  # Duration of audio cue
                amplitude=0.9,  # Volume, from 0.0 to 1.0
                sample_rate=sample_rate,
            )

            # Create a sound object from the numpy array.
            pure_tone_start = pygame.sndarray.make_sound(tone_data_start)
            pure_tone_end = pygame.sndarray.make_sound(tone_data_end)

        # ------------------------------------------------------------------------------
        # *** Prepare visual stimulus generation

        # Get screen dimensions and set up full screen.
        screen_info = pygame.display.Info()
        screen_width = screen_info.current_w
        screen_height = screen_info.current_h
        screen = pygame.display.set_mode(
            (screen_width, screen_height), pygame.FULLSCREEN
        )
        pygame.display.set_caption("Silent Reading Experiment")
        pygame.mouse.set_visible(False)

        try:
            # Initial grey screen.
            pygame.time.wait(100)
            screen.fill(text_config.rest_condition_color)
            pygame.display.flip()
            pygame.time.wait(100)
            screen.fill(text_config.rest_condition_color)
            pygame.display.flip()

            stimulus_fonts = construct_fonts(font_sizes=stimulus_font_sizes)

            # Pause for specified number of milliseconds.
            pygame.time.delay(int(round(initial_rest_duration * 1000.0)))

            # Count stimuli to introduce breaks between blocks.
            stimulus_block_counter = 0

            # Loop through words.
            for word, is_target_event in zip(text, is_target):
                if not running:  # Check for quit event
                    break

                # Randomly sample a font (we render the stimulus using different fonts
                # to achieve different stimulus appearance in terms of low-level visual
                # features.
                font_data = random.choice(stimulus_fonts)
                font_name = font_data["font_name"]
                font_size = font_data["font_size"]
                font_is_bold = font_data["font_is_bold"]
                font_is_italic = font_data["font_is_italic"]
                font_color = random.choice(text_config.font_colors)
                font_spacing = np.random.uniform(
                    low=stimulus_font_min_spacing,
                    high=stimulus_font_max_spacing,
                )

                stimulus_text = render_spaced_text(
                    text=word,
                    font=font_data["font"],
                    color=font_color,
                    spacing=font_spacing,
                )

                stimulus_rect = stimulus_text.get_rect(
                    center=(screen_width // 2, screen_height // 2)
                )
                screen.blit(stimulus_text, stimulus_rect)
                pygame.display.flip()
                # Start of stimulus presentation.
                t_stim_start = time()

                response_made = False
                response_time = np.nan
                response_deadline = t_stim_start + response_window_duration

                # Wait for stimulus duration, but check for responses continuously.
                t_stim_end_expected = t_stim_start + stimulus_duration
                while time() < t_stim_end_expected:
                    for event in pygame.event.get():
                        if event.type == pygame.QUIT:
                            running = False
                        if event.type == pygame.KEYDOWN:
                            keydown_time = time()
                            if event.key == pygame.K_ESCAPE:
                                running = False
                            # Check for space bar press within the response window.
                            if event.key == pygame.K_SPACE and not response_made:
                                if keydown_time < response_deadline:
                                    response_made = True
                                    response_time = keydown_time - t_stim_start
                                    print(f"Response time: {round(response_time, 3)}")
                                    if is_target_event:
                                        # Hit.
                                        n_hits += 1
                                    else:
                                        # False alarm.
                                        n_false_alarms += 1
                    if not running:
                        break
                if not running:
                    break

                # End of stimulus presentation. Display ISI grey screen.
                screen.fill(text_config.rest_condition_color)
                pygame.display.flip()
                t_stim_end_actual = time()

                # Time until when to show grey screen (ISI).
                t_isi_end = (
                    t_stim_end_actual
                    + isi_duration
                    + np.random.uniform(low=0.0, high=isi_jitter)
                )

                if is_target_event:
                    # If this is a target event, prolong the ISI duration, to allow the
                    # subject to respond before the onset of the next stimulus, to
                    # reduce the probability of a motor response artefact in the
                    # subsequent trial.
                    t_isi_end += isi_extension_target

                # Continue checking for late responses or quit events.
                while time() < t_isi_end:
                    for event in pygame.event.get():
                        if event.type == pygame.QUIT:
                            running = False
                        if event.type == pygame.KEYDOWN:
                            keydown_time = time()
                            if event.key == pygame.K_ESCAPE:
                                running = False
                            # Still check for spacebar presses that are within the
                            # response window for target events.
                            if event.key == pygame.K_SPACE and not response_made:
                                if keydown_time < response_deadline:
                                    response_made = True
                                    response_time = keydown_time - t_stim_start
                                    print(f"Response time: {round(response_time, 3)}")
                                    if is_target_event:
                                        # Hit.
                                        n_hits += 1
                                    else:
                                        # False alarm.
                                        n_false_alarms += 1
                    if not running:
                        break
                if not running:
                    break

                if not running:
                    break

                stimulus_block_counter += 1

                if stimulus_block_counter == stimuli_per_block:
                    # Inter-block grey screen.
                    screen.fill(text_config.rest_condition_color)
                    pygame.display.flip()
                    # Start of inter-block interval.
                    t_ibi_start = time()

                    if use_ibi_audio_cue:
                        # Audio cue to signal the beginning of the inter-block interval.
                        pure_tone_start.play()

                    # End of inter-block interval.
                    t_ibi_end = t_ibi_start + inter_block_rest_duration

                    if use_ibi_audio_cue:
                        # Time when to play audio cue to signal end of inter-block
                        # interval.
                        t_ibi_end_audio_cue = t_ibi_end - pure_tone_end_delay
                        ibi_end_cue_played_yet = False

                    while time() < t_ibi_end:  # Continue inter-block interval?
                        if (
                            use_ibi_audio_cue
                            and (t_ibi_end_audio_cue <= time())
                            and not ibi_end_cue_played_yet
                        ):  # Time to play end of inter-block interval cue?
                            # Play the cue to signal the end of the inter-block
                            # interval.
                            pure_tone_end.play()
                            ibi_end_cue_played_yet = True

                    stimulus_block_counter = 0

            # End of word loop.

            # Calculate behavioural results.
            n_misses = n_total_targets - n_hits

            if running:
                # Display behavioural results.
                screen.fill(text_config.rest_condition_color)

                # Behavioural results title.
                title_font = pygame.font.Font(None, 72)
                title_text = title_font.render(
                    "Experiment Complete",
                    True,
                    text_config.font_colors[0],
                )
                title_rect = title_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 - 150)
                )
                screen.blit(title_text, title_rect)

                # Behavioural results text.
                results_font = pygame.font.Font(None, 56)
                hits_text = results_font.render(
                    f"Hits: {n_hits}",
                    True,
                    text_config.font_colors[0],
                )
                misses_text = results_font.render(
                    f"Misses: {n_misses}",
                    True,
                    text_config.font_colors[0],
                )
                false_alarms_text = results_font.render(
                    f"False Alarms: {n_false_alarms}",
                    True,
                    text_config.font_colors[0],
                )

                # Position and display results
                hits_rect = hits_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 - 20)
                )
                misses_rect = misses_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 + 40)
                )
                false_alarms_rect = false_alarms_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 + 100)
                )

                screen.blit(hits_text, hits_rect)
                screen.blit(misses_text, misses_rect)
                screen.blit(false_alarms_text, false_alarms_rect)

                pygame.display.flip()
                pygame.time.wait(5000)  # Show results for 5 seconds

            running = False

        except Exception as e:
            print(f"An error occurred during the experiment: {e}")
            print(traceback.format_exc())
            running = False
        finally:
            pygame.quit()
            print("Experiment closed.")



def map_session_config_to_experiment_config(
    *,
    session_config: dict,
    experiment_config: dict,
):
    """
    Map session config to experiment config. Specifically:
    - subject (integer), e.g. 1 -> subject_id (string), e.g. "sub-001"
    - session (integer), e.g. 1 -> session_id (string), e.g. "session-001"
    - run (integer), e.g. 1 -> word_idx_start (intger) = n_words_to_show * (run - 1)
    """
    subject = session_config["subject"]
    session = session_config["session"]
    run = session_config["run"]

    subject_id = f"sub-{subject:03}"
    session_id = f"session-{session:03}"

    n_words_to_show = experiment_config["n_words_to_show"]
    # Minus one because we start at word zero on first run.
    word_idx_start = n_words_to_show * (run - 1)

    experiment_config["subject_id"] = subject_id
    experiment_config["session_id"] = session_id
    experiment_config["word_idx_start"] = word_idx_start

    return experiment_config



import json
from time import time

import h5py
import numpy as np

from nubrain.experiment_image.image_config import ImageConfig
from nubrain.image.tools import load_image_as_bytes, resize_image

image_config = ImageConfig()


def eeg_data_logging(subprocess_params: dict):
    """
    Log experimental data. Save to local hdf file. To be run in separate process (using
    multiprocessing).
    """
    # ----------------------------------------------------------------------------------
    # *** Get parameters

    image_directory = subprocess_params["image_directory"]

    subject_id = subprocess_params["subject_id"]
    session_id = subprocess_params["session_id"]

    # EEG parameters
    eeg_board_description = subprocess_params["eeg_board_description"]
    eeg_sampling_rate = subprocess_params["eeg_sampling_rate"]
    n_channels_total = subprocess_params["n_channels_total"]
    eeg_channels = subprocess_params["eeg_channels"]
    marker_channel = subprocess_params["marker_channel"]
    eeg_channel_mapping = subprocess_params["eeg_channel_mapping"]
    eeg_device_address = subprocess_params["eeg_device_address"]

    # Timing parameters
    initial_rest_duration = subprocess_params["initial_rest_duration"]
    image_duration = subprocess_params["image_duration"]
    isi_duration = subprocess_params["isi_duration"]
    inter_block_grey_duration = subprocess_params["inter_block_grey_duration"]

    # Experiment structure
    n_blocks = subprocess_params["n_blocks"]
    images_per_block = subprocess_params["images_per_block"]

    utility_frequency = subprocess_params["utility_frequency"]

    # nubrain_endpoint = subprocess_params["nubrain_endpoint"]
    # nubrain_api_key = subprocess_params["nubrain_api_key"]

    path_out_data = subprocess_params["path_out_data"]

    data_logging_queue = subprocess_params["data_logging_queue"]

    # ----------------------------------------------------------------------------------
    # *** Create and initialize HDF5 file

    experiment_metadata = {
        "config_version": image_config.config_version,
        "subject_id": subject_id,
        "session_id": session_id,
        "image_directory": image_directory,
        "rest_condition_color": image_config.rest_condition_color,
        "stim_start_marker": image_config.stim_start_marker,
        "stim_end_marker": image_config.stim_end_marker,
        "hdf5_dtype": image_config.hdf5_dtype,
        "max_img_storage_dimension": image_config.max_img_storage_dimension,
        "experiment_start_time": time(),
        # EEG parameters
        "eeg_board_description": eeg_board_description,
        "eeg_sampling_rate": eeg_sampling_rate,
        "n_channels_total": n_channels_total,
        "eeg_channel_mapping": eeg_channel_mapping,
        # Timing parameters
        "initial_rest_duration": initial_rest_duration,
        "image_duration": image_duration,
        "isi_duration": isi_duration,
        "inter_block_grey_duration": inter_block_grey_duration,
        # Experiment structure
        "n_blocks": n_blocks,
        "images_per_block": images_per_block,
        # Misc
        "utility_frequency": utility_frequency,
    }

    # Parameters not used by DSI-24, for compatibility with Cyton board.
    if eeg_device_address is not None:
        experiment_metadata["eeg_device_address"] = eeg_device_address
    if eeg_channels is not None:
        experiment_metadata["eeg_channels"] = eeg_channels
    if marker_channel is not None:
        experiment_metadata["marker_channel"] = marker_channel

    print(f"Initializing HDF5 file at: {path_out_data}")
    with h5py.File(path_out_data, "w") as file:
        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for metadata

        # Create group for metadata.
        metadata_group = file.create_group("metadata")

        # Iterate over the Python dictionary and save each item as an attribute of the
        # "metadata" group.
        for key, value in experiment_metadata.items():
            # HDF5 attributes have limitations on data types. Complex types like
            # dictionaries or tuples are not natively supported. We check if the value
            # is a type that needs to be converted to a string. JSON is a convenient
            # format for this serialization.
            if isinstance(value, (dict, list, tuple)):
                # Serialize the complex type into a JSON string.
                metadata_group.attrs[key] = json.dumps(value)
            else:
                metadata_group.attrs[key] = value

        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for EEG data

        # Initialize dataset for EEG and additional channels. To handle a variable
        # number of timesteps, create a resizable dataset. We specify an initial shape
        # but set the 'maxshape' to allow one of the dimensions to be unlimited (by
        # setting it to None). 'chunks=True' is recommended for resizable datasets for
        # better performance. It lets h5py decide the chunk size.

        file.create_dataset(
            "eeg_data",
            shape=(n_channels_total, 0),
            maxshape=(n_channels_total, None),  # fixed_channels, unlimited_timesteps
            dtype=image_config.hdf5_dtype,
            chunks=True,
        )

        file.create_dataset(
            "eeg_timestamps",
            shape=(0,),
            maxshape=(None,),
            dtype="float64",  # LSL timestamps
            chunks=True,
        )

        file.create_dataset(
            "marker_data",
            shape=(2, 0),  # timestamp, marker value
            maxshape=(2, None),
            dtype="float64",
            chunks=True,
        )

        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for stimulus data

        # Define the compound datatype for stimulus data. This is like defining the
        # columns of a table. Use a special vlen dtype for the variable-sized image
        # data.
        stimulus_dtype = np.dtype(
            [
                ("stimulus_start_time", np.float64),
                ("stimulus_end_time", np.float64),
                ("stimulus_duration_s", np.float64),
                ("image_file_path", h5py.string_dtype(encoding="utf-8")),
                ("image_category", h5py.string_dtype(encoding="utf-8")),
                # ("image_description", h5py.string_dtype(encoding="utf-8")),
                (
                    "image_bytes",
                    h5py.vlen_dtype(np.uint8),
                ),  # For variable-length byte arrays
                ("is_target_event", np.bool),
                ("response_time_s", np.float64),
            ]
        )

        n_images = n_blocks * images_per_block

        file.create_dataset(
            "stimulus_data",
            (n_images,),
            dtype=stimulus_dtype,
        )

        # ------------------------------------------------------------------------------
        # *** Initialize hdf5 dataset for behavioural data

        behavioural_dtype = np.dtype(
            [
                ("n_total_targets", np.int64),
                ("n_hits", np.int64),
                ("n_misses", np.int64),
                ("n_false_alarms", np.int64),
            ]
        )

        file.create_dataset(
            "behavioural_data",
            (1,),
            dtype=behavioural_dtype,
        )

    # ----------------------------------------------------------------------------------
    # *** Experiment loop

    stimulus_counter = 0

    while True:
        new_data = data_logging_queue.get(block=True)

        if new_data is None:
            # Received None. End process.
            print("Ending preprocessing & data saving process.")
            break

        data_type = new_data["type"]

        with h5py.File(path_out_data, "a") as file:
            # --------------------------------------------------------------------------
            # *** Write EEG data to hdf5 file

            if data_type == "eeg":
                new_eeg_data = new_data.get("eeg_data")
                new_timestamps = new_data.get("eeg_timestamps")

                if new_eeg_data is not None and new_eeg_data.size > 0:
                    # Write EEG data.
                    hdf5_eeg_data = file["eeg_data"]
                    n_existing = hdf5_eeg_data.shape[1]
                    n_new = new_eeg_data.shape[1]
                    hdf5_eeg_data.resize(n_existing + n_new, axis=1)
                    hdf5_eeg_data[:, n_existing:] = new_eeg_data

                    # Write EEG timestamps.
                    hdf5_timestamps = file["eeg_timestamps"]
                    n_existing_ts = hdf5_timestamps.shape[0]
                    hdf5_timestamps.resize(n_existing_ts + n_new, axis=0)
                    hdf5_timestamps[n_existing_ts:] = new_timestamps

            # --------------------------------------------------------------------------
            # *** Write stimulus markers to hdf5 file

            elif data_type == "marker":
                marker_value = new_data.get("marker_value")
                marker_timestamp = new_data.get("timestamp")

                if marker_value is not None:
                    hdf5_marker_data = file["marker_data"]
                    n_existing = hdf5_marker_data.shape[1]
                    hdf5_marker_data.resize(n_existing + 1, axis=1)
                    hdf5_marker_data[:, n_existing] = (marker_timestamp, marker_value)

            # --------------------------------------------------------------------------
            # *** Write stimulus data to hdf5 file

            elif data_type == "stimulus":
                new_stimulus_data = new_data.get("stimulus_data")

                if new_stimulus_data is not None:
                    hdf5_stimulus_data = file["stimulus_data"]

                    image_file_path = new_stimulus_data["image_file_path"]
                    image_bytes = load_image_as_bytes(image_path=image_file_path)
                    image_bytes = resize_image(image_bytes=image_bytes)

                    stimulus_start_time = new_stimulus_data["stimulus_start_time"]
                    stimulus_end_time = new_stimulus_data["stimulus_end_time"]
                    stimulus_duration_s = new_stimulus_data["stimulus_duration_s"]
                    image_file_path = new_stimulus_data["image_file_path"]
                    image_category = new_stimulus_data["image_category"]
                    # image_description = new_stimulus_data["image_description"]
                    is_target_event = new_stimulus_data["is_target_event"]
                    response_time_s = new_stimulus_data["response_time_s"]

                    data_to_write = np.empty((1,), dtype=stimulus_dtype)
                    data_to_write[0]["stimulus_start_time"] = stimulus_start_time
                    data_to_write[0]["stimulus_end_time"] = stimulus_end_time
                    data_to_write[0]["stimulus_duration_s"] = stimulus_duration_s
                    data_to_write[0]["image_file_path"] = image_file_path
                    data_to_write[0]["image_category"] = image_category
                    # data_to_write[0]["image_description"] = image_description
                    data_to_write[0]["is_target_event"] = is_target_event
                    data_to_write[0]["response_time_s"] = response_time_s
                    # The image data is stored as a numpy array of bytes (uint8).
                    data_to_write[0]["image_bytes"] = np.frombuffer(
                        image_bytes,
                        dtype=np.uint8,
                    )

                    # Write the structured array to the dataset.
                    hdf5_stimulus_data[stimulus_counter] = data_to_write

                    print(f"Stimulus counter: {stimulus_counter}")
                    stimulus_counter += 1

            # --------------------------------------------------------------------------
            # *** Write behavioural data to hdf5 file

            elif data_type == "behavioural":
                new_behavioural_data = new_data.get("behavioural_data")

                if new_behavioural_data is not None:
                    hdf5_behavioural_data = file["behavioural_data"]

                    n_total_targets = new_behavioural_data["n_total_targets"]
                    n_hits = new_behavioural_data["n_hits"]
                    n_misses = new_behavioural_data["n_misses"]
                    n_false_alarms = new_behavioural_data["n_false_alarms"]

                    data_to_write = np.empty((1,), dtype=behavioural_dtype)

                    data_to_write[0]["n_total_targets"] = n_total_targets
                    data_to_write[0]["n_hits"] = n_hits
                    data_to_write[0]["n_misses"] = n_misses
                    data_to_write[0]["n_false_alarms"] = n_false_alarms

                    # Write the structured array to the dataset.
                    hdf5_behavioural_data[0] = data_to_write


# End of data preprocessing process.



from nubrain.global_config import GlobalConfig


class TextConfig:
    def __init__(self, version: str = "v1"):
        global_config = GlobalConfig()

        self.config_version = version
        # Color values for experimental rest condition (e.g. grey).
        self.rest_condition_color = (0, 0, 0)
        # Font colors. If more than one, will sample randomly on each trial. Will use
        # first color in list for rendering behavioural results.
        self.font_colors = [
            (255, 255, 255),
            (255, 0, 0),
            (0, 255, 0),
            (0, 0, 255),
            (0, 255, 255),
            (255, 0, 255),
            (255, 255, 0),
        ]
        # Markers for stimulus start and end (will be stored in marker channel).
        self.stim_start_marker = global_config.stim_start_marker
        self.stim_end_marker = global_config.stim_end_marker
        # Data type for EEG data to use when saving to hdf5 file.
        self.hdf5_dtype = global_config.hdf5_dtype



import multiprocessing as mp
import os
import random
import traceback
from time import sleep, time

import numpy as np
import pygame

from nubrain.device.device_interface import create_eeg_device
from nubrain.experiment_image.data import eeg_data_logging
from nubrain.experiment_image.image_config import ImageConfig
from nubrain.experiment_image.randomize_conditions import (
    create_balanced_list,
    sample_next_image,
    shuffle_with_repetitions,
)
from nubrain.image.tools import get_all_images, load_and_scale_image
from nubrain.misc.datetime import get_formatted_current_datetime

mp.set_start_method("spawn", force=True)  # Necessary on if running on windows?


def experiment_image(config: dict):
    # ----------------------------------------------------------------------------------
    # *** Get config

    device_type = config["device_type"]
    lsl_stream_name = config.get("lsl_stream_name", "DSI-24")

    subject_id = config["subject_id"]
    session_id = config["session_id"]

    output_directory = config["output_directory"]
    image_directory = config["image_directory"]

    eeg_channel_mapping = config.get("eeg_channel_mapping", None)

    utility_frequency = config["utility_frequency"]

    initial_rest_duration = config["initial_rest_duration"]
    image_duration = config["image_duration"]
    isi_duration = config["isi_duration"]
    isi_jitter = config["isi_jitter"]
    inter_block_grey_duration = config["inter_block_grey_duration"]

    n_blocks = config["n_blocks"]
    images_per_block = config["images_per_block"]
    n_target_events = config["n_target_events"]

    response_window_duration = config["response_window_duration"]

    eeg_device_address = config.get("eeg_device_address", None)

    image_config = ImageConfig()

    # ----------------------------------------------------------------------------------
    # *** Test if output path exists

    if not os.path.isdir(output_directory):
        raise AssertionError(f"Target directory does not exist: {output_directory}")

    current_datetime = get_formatted_current_datetime()
    path_out_data = os.path.join(output_directory, f"eeg_{current_datetime}.h5")

    if os.path.isfile(path_out_data):
        raise AssertionError(f"Target file already exists: {path_out_data}")

    # ----------------------------------------------------------------------------------
    # *** Get input images & their categories

    images_and_categories = get_all_images(image_directory=image_directory)

    if not images_and_categories:
        raise AssertionError(f"Found no images at {image_directory}")
    print(f"Found {len(images_and_categories)} images")

    # ----------------------------------------------------------------------------------
    # *** Create pseudo-random condition order

    # List with all unique image categories (e.g. `["apple", "banana", ...]`).
    image_categories = list(set([x["image_category"] for x in images_and_categories]))

    n_trials = n_blocks * images_per_block

    # Order of image categories.
    trial_order = create_balanced_list(
        image_categories=image_categories,
        target_length=n_trials,
    )
    random.shuffle(trial_order)

    trial_order = shuffle_with_repetitions(
        list_with_duplicates=trial_order,
        repetitions=n_target_events,
        minimize_runs=True,
    )

    # Mapping from image categories to image file paths, e.g. `{"apple":
    # ["/path/to/apple_1.png", "/path/to/apple_2.png", ...], "banana":
    # ["/path/to/banana_2.png", ...]}`.
    category_to_filepath = {}
    for item in images_and_categories:
        image_category = item["image_category"]
        image_filepath = item["image_file_path"]
        if image_category in category_to_filepath:
            category_to_filepath[image_category].append(image_filepath)
        else:
            category_to_filepath[image_category] = [image_filepath]

    previous_image_file_path = None
    previous_image_category = None

    # ----------------------------------------------------------------------------------
    # *** Prepare EEG measurement

    print(f"Initializing EEG device: {device_type}")

    device_kwargs = {"eeg_channel_mapping": eeg_channel_mapping}
    if device_type in ["cyton", "synthetic"]:
        device_kwargs["eeg_device_address"] = eeg_device_address
    elif device_type == "dsi24":
        device_kwargs["lsl_stream_name"] = lsl_stream_name
    else:
        raise ValueError(f"Unexpected `device_type`: {device_type}")

    eeg_device = create_eeg_device(device_type, **device_kwargs)

    eeg_device.prepare_session()

    # This is a bit clunky. At this point, `eeg_channel_mapping` is None or a dict with
    # a channel mapping from the config yaml file. Overwrite it with the channel mapping
    # from the device (in case of the DSI-24 device, the channel mapping from the device
    # is used in any case).
    eeg_channel_mapping = eeg_device.eeg_channel_mapping

    # Need to start the stream before calling `eeg_device.get_device_info()`, because
    # we retrieve data from board to determine data shape (number of channels).
    eeg_device.start_stream()
    sleep(0.1)

    # Get device info.
    device_info = eeg_device.get_device_info()
    eeg_board_description = device_info["board_description"]
    eeg_sampling_rate = device_info["sampling_rate"]
    eeg_channels = device_info["eeg_channels"]
    marker_channel = device_info["marker_channel"]
    n_channels_total = device_info["n_channels_total"]

    if device_type in ["cyton", "synthetic"]:
        # For Cyton device, we need to get the number of EEG channels from the device
        # (not sure, this might only work after starting the stream).
        eeg_device.eeg_channels = eeg_channels
        eeg_device.timestamp_channel = eeg_board_description["timestamp_channel"]

    print(f"Board: {eeg_board_description['name']}")
    print(f"Sampling Rate: {eeg_sampling_rate} Hz")
    print(f"EEG Channels: {eeg_channels}")
    print(f"Marker Channel: {marker_channel}")
    print(f"EEG Channel Mapping: {eeg_channel_mapping}")

    board_data, board_timestamps = eeg_device.get_board_data()

    print(f"Board data dtype: {board_data.dtype}")
    print(f"Board data shape: {board_data.shape}")
    print(f"Board timestamps shape: {board_timestamps.shape}")

    # ----------------------------------------------------------------------------------
    # *** Start data logging subprocess

    data_logging_queue = mp.Queue()

    subprocess_params = {
        "device_type": device_type,
        "subject_id": subject_id,
        "session_id": session_id,
        "image_directory": image_directory,
        # EEG parameters
        "eeg_board_description": eeg_board_description,
        "eeg_sampling_rate": eeg_sampling_rate,
        "n_channels_total": n_channels_total,
        "eeg_channels": eeg_channels,
        "marker_channel": marker_channel,
        "eeg_channel_mapping": eeg_channel_mapping,
        "eeg_device_address": eeg_device_address,
        # Timing parameters
        "initial_rest_duration": initial_rest_duration,
        "image_duration": image_duration,
        "isi_duration": isi_duration,
        "inter_block_grey_duration": inter_block_grey_duration,
        # Experiment structure
        "n_blocks": n_blocks,
        "images_per_block": images_per_block,
        # Misc
        "utility_frequency": utility_frequency,
        "path_out_data": path_out_data,
        "data_logging_queue": data_logging_queue,
    }

    logging_process = mp.Process(target=eeg_data_logging, args=(subprocess_params,))
    logging_process.daemon = True
    logging_process.start()

    # ----------------------------------------------------------------------------------
    # *** Start experiment

    # Performance counters.
    n_hits = 0
    n_false_alarms = 0
    n_total_targets = 0

    running = True
    while running:
        pygame.init()

        # Get screen dimensions and set up full screen.
        screen_info = pygame.display.Info()
        screen_width = screen_info.current_w
        screen_height = screen_info.current_h
        screen = pygame.display.set_mode(
            (screen_width, screen_height), pygame.FULLSCREEN
        )
        pygame.display.set_caption("Image Presentation Experiment")
        pygame.mouse.set_visible(False)

        idx_trial = 0

        try:
            # Initial grey screen.
            pygame.time.wait(100)
            screen.fill(image_config.rest_condition_color)
            pygame.display.flip()
            pygame.time.wait(100)
            screen.fill(image_config.rest_condition_color)
            pygame.display.flip()

            # Clear board buffer.
            _, _ = eeg_device.get_board_data()

            # Pause for specified number of milliseconds.
            pygame.time.delay(int(round(initial_rest_duration * 1000.0)))

            # Block loop.
            for idx_block in range(n_blocks):
                # Image loop (within a block).
                for image_count in range(images_per_block):
                    if not running:  # Check for quit event
                        break

                    # Sample the next image.
                    next_image_category = trial_order[idx_trial]
                    next_image_file_path = sample_next_image(
                        next_image_category=next_image_category,
                        category_to_filepath=category_to_filepath,
                        previous_image_file_path=previous_image_file_path,
                    )

                    # Load the next image.
                    image_and_metadata = None
                    while image_and_metadata is None:
                        image_and_metadata = load_and_scale_image(
                            image_file_path=next_image_file_path,
                            screen_width=screen_width,
                            screen_height=screen_height,
                        )

                    current_image = image_and_metadata["image"]

                    img_rect = current_image.get_rect(
                        center=(screen_width // 2, screen_height // 2)
                    )
                    screen.fill(image_config.rest_condition_color)
                    screen.blit(current_image, img_rect)
                    pygame.display.flip()
                    t_stim_start = time()  # Start of stimulus presentation.

                    # Insert stimulus start marker and get its timestamp.
                    marker_val, marker_ts = eeg_device.insert_marker(
                        image_config.stim_start_marker
                    )
                    if marker_val is not None:
                        data_logging_queue.put(
                            {
                                "type": "marker",
                                "marker_value": marker_val,
                                "timestamp": marker_ts,
                            }
                        )

                    # Send pre-stimulus EEG data (to avoid buffer overflow).
                    eeg_data, eeg_ts = eeg_device.get_board_data()
                    if eeg_data.size > 0:
                        data_logging_queue.put(
                            {
                                "type": "eeg",
                                "eeg_data": eeg_data,
                                "eeg_timestamps": eeg_ts,
                            }
                        )

                    # Determine if the current trial is a target event.
                    is_target_event = False
                    if idx_trial > 0 and next_image_category == previous_image_category:
                        is_target_event = True
                        n_total_targets += 1

                    response_made = False
                    response_time = np.nan
                    response_deadline = t_stim_start + response_window_duration

                    # Wait for image duration, but check for responses continuously.
                    t_stim_end_expected = t_stim_start + image_duration
                    while time() < t_stim_end_expected:
                        for event in pygame.event.get():
                            if event.type == pygame.QUIT:
                                running = False
                            if event.type == pygame.KEYDOWN:
                                keydown_time = time()
                                if event.key == pygame.K_ESCAPE:
                                    running = False
                                # Check for space bar press within the response window.
                                if event.key == pygame.K_SPACE and not response_made:
                                    if keydown_time < response_deadline:
                                        response_made = True
                                        response_time = keydown_time - t_stim_start
                                        print(
                                            f"Response time: {round(response_time, 3)}"
                                        )
                                        if is_target_event:
                                            # Hit.
                                            n_hits += 1
                                        else:
                                            # False alarm.
                                            n_false_alarms += 1
                        if not running:
                            break
                    if not running:
                        break

                    # End of stimulus presentation. Display ISI grey screen.
                    screen.fill(image_config.rest_condition_color)
                    pygame.display.flip()
                    t_stim_end_actual = time()

                    marker_val, marker_ts = eeg_device.insert_marker(
                        image_config.stim_end_marker
                    )
                    if marker_val is not None:
                        data_logging_queue.put(
                            {
                                "type": "marker",
                                "marker_value": marker_val,
                                "timestamp": marker_ts,
                            }
                        )

                    eeg_data, eeg_ts = eeg_device.get_board_data()
                    if eeg_data.size > 0:
                        data_logging_queue.put(
                            {
                                "type": "eeg",
                                "eeg_data": eeg_data,
                                "eeg_timestamps": eeg_ts,
                            }
                        )

                    # Time until when to show grey screen (ISI).
                    t_isi_end = (
                        t_stim_end_actual
                        + isi_duration
                        + np.random.uniform(low=0.0, high=isi_jitter)
                    )

                    # Continue checking for late responses or quit events.
                    while time() < t_isi_end:
                        for event in pygame.event.get():
                            if event.type == pygame.QUIT:
                                running = False
                            if event.type == pygame.KEYDOWN:
                                keydown_time = time()
                                if event.key == pygame.K_ESCAPE:
                                    running = False
                                # Still check for spacebar presses that are within the
                                # response window for target events.
                                if event.key == pygame.K_SPACE and not response_made:
                                    if keydown_time < response_deadline:
                                        response_made = True
                                        response_time = keydown_time - t_stim_start
                                        print(
                                            f"Response time: {round(response_time, 3)}"
                                        )
                                        if is_target_event:
                                            # Hit.
                                            n_hits += 1
                                        else:
                                            # False alarm.
                                            n_false_alarms += 1
                        if not running:
                            break
                    if not running:
                        break

                    stimulus_data = {
                        "stimulus_start_time": t_stim_start,
                        "stimulus_end_time": t_stim_end_actual,
                        "stimulus_duration_s": t_stim_end_actual - t_stim_start,
                        "image_file_path": next_image_file_path,
                        "image_category": next_image_category,
                        "is_target_event": is_target_event,
                        "response_time_s": response_time,
                    }
                    data_logging_queue.put(
                        {"type": "stimulus", "stimulus_data": stimulus_data}
                    )

                    # Update tracking variables for the next loop iteration.
                    previous_image_file_path = next_image_file_path
                    previous_image_category = next_image_category
                    idx_trial += 1

                if not running:
                    break

                # Send post-stimulus EEG data (to avoid buffer overflow).
                eeg_data, eeg_ts = eeg_device.get_board_data()
                if eeg_data.size > 0:
                    data_logging_queue.put(
                        {"type": "eeg", "eeg_data": eeg_data, "eeg_timestamps": eeg_ts}
                    )

                # Inter-block grey screen.
                screen.fill(image_config.rest_condition_color)
                pygame.display.flip()
                # We already waited for the ISI duration, therefore subtract it from the
                # inter block duration. Avoid negative value in case ISI duration is
                # longer than inter block duration.
                remaining_wait = max((inter_block_grey_duration - isi_duration), 0.0)
                pygame.time.delay(int(round(remaining_wait * 1000.0)))

            # Calculate behavioural results.
            n_misses = n_total_targets - n_hits

            # Write behavioural results to hdf5 file.
            behavioural_data = {
                "n_total_targets": n_total_targets,
                "n_hits": n_hits,
                "n_misses": n_misses,
                "n_false_alarms": n_false_alarms,
            }
            data_logging_queue.put(
                {"type": "behavioural", "behavioural_data": behavioural_data}
            )

            if running:
                # Display behavioural results.
                screen.fill(image_config.rest_condition_color)

                # Behavioural results title.
                title_font = pygame.font.Font(None, 72)
                title_text = title_font.render("Experiment Complete", True, (0, 0, 0))
                title_rect = title_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 - 150)
                )
                screen.blit(title_text, title_rect)

                # Behavioural results text.
                results_font = pygame.font.Font(None, 56)
                hits_text = results_font.render(f"Hits: {n_hits}", True, (0, 0, 0))
                misses_text = results_font.render(
                    f"Misses: {n_misses}", True, (0, 0, 0)
                )
                false_alarms_text = results_font.render(
                    f"False Alarms: {n_false_alarms}", True, (0, 0, 0)
                )

                # Position and display results
                hits_rect = hits_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 - 20)
                )
                misses_rect = misses_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 + 40)
                )
                false_alarms_rect = false_alarms_text.get_rect(
                    center=(screen_width // 2, screen_height // 2 + 100)
                )

                screen.blit(hits_text, hits_rect)
                screen.blit(misses_text, misses_rect)
                screen.blit(false_alarms_text, false_alarms_rect)

                pygame.display.flip()
                pygame.time.wait(5000)  # Show results for 5 seconds

            running = False

            # Send final board data.
            eeg_data, eeg_ts = eeg_device.get_board_data()
            if eeg_data.size > 0:
                data_logging_queue.put(
                    {"type": "eeg", "eeg_data": eeg_data, "eeg_timestamps": eeg_ts}
                )

        except Exception as e:
            print(f"An error occurred during the experiment: {e}")
            print(traceback.format_exc())
            running = False
        finally:
            pygame.quit()
            print("Experiment closed.")

    eeg_device.stop_stream()
    eeg_device.release_session()

    print("Join process for sending data")
    data_logging_queue.put(None)
    logging_process.join()



import argparse
import os

from nubrain.experiment_image.load_config import load_config_image_yaml
from nubrain.experiment_text.demo import text_demo
from nubrain.experiment_text.gui import SessionConfigEditor
from nubrain.experiment_text.load_experiment_config import load_config_text_yaml
from nubrain.experiment_text.map_config import map_session_config_to_experiment_config

# Wrap these imports in try, so that the other modules can be imported without
# dependency on pylsl for demo mode.
try:
    from nubrain.experiment_eeg_to_image_v1.load_config import (
        load_config_yaml_eeg_to_image_v1,
    )
    from nubrain.experiment_eeg_to_image_v1.main import experiment_eeg_to_image_v1
    from nubrain.experiment_eeg_to_image_v1.main_autoregressive import (
        experiment_eeg_to_image_v1_autoregressive,
    )
    from nubrain.experiment_image.main import experiment_image
    from nubrain.experiment_text.main import experiment_text
except Exception as e:
    experiment_image = None
    experiment_text = None
    load_config_yaml_eeg_to_image_v1 = None
    experiment_eeg_to_image_v1 = None
    experiment_eeg_to_image_v1_autoregressive = None
    print(f"Failed to import nubrain main module: {e}")

from nubrain.live_demo.main import run_live_demo


def main():
    """
    Main entry point for the nubrain command-line application.
    """
    # Initialize the parser.
    parser = argparse.ArgumentParser(description="nubrain command-line interface.")

    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Path to the experiment configuration YAML file.",
    )

    # Which experimental mode to use. Options:
    # - "data_collection_image": Data collection mode for image stimuli.
    # - "demo_text":  Demo mode with text stimuli for instructing participants.
    # - "data_collection_text":  Data collection mode for text stimuli.
    # - "eeg_to_image":  After presenting each image, directly reconstruct the image,
    #   then show the next image.
    # - "eeg_to_image_autoregressive": After presenting an image, directly reconstruct
    #   the image, and then show the reconstructed image as the next stimulus.
    # - "eeg_to_image_live_demo": Use cache.
    parser.add_argument(
        "--mode",
        type=str,
        default="data_collection_text",
        help="Which experimental mode to use",
    )

    args = parser.parse_args()

    print("nubrain")
    print(f"Configuration file provided: {args.config}")

    input_file_path = args.config

    mode = args.mode

    # Load EEG experiment config from yaml file.
    if mode == "data_collection_image":
        # Data collection mode, image stimuli.
        config = load_config_image_yaml(yaml_file_path=input_file_path)
    elif mode in ["demo_text", "data_collection_text"]:
        # Data collection mode, text stimuli.
        config = load_config_text_yaml(yaml_file_path=input_file_path)
    elif mode in ["eeg_to_image", "eeg_to_image_autoregressive"]:
        # Live EEG to image generation mode. Use corresponding config file loading
        # function (different parameters than regular data collection).
        config = load_config_yaml_eeg_to_image_v1(yaml_file_path=input_file_path)
    elif mode == "eeg_to_image_live_demo":
        pass
    else:
        raise AssertionError(f"Unknown experimental mode: {mode}")

    # Run experiment.
    if mode == "data_collection_image":
        experiment_image(config=config)
    elif mode in ["demo_text", "data_collection_text"]:
        # Show GUI for user to update session parameters (subject ID, next run).
        session_config_path = os.path.join(
            os.path.dirname(__file__),
            "experiment_text",
            "session_config.yaml",
        )
        gui = SessionConfigEditor(session_config_path=session_config_path)
        session_config = gui.run()
        if session_config is None:
            # User pressed the cancel button.
            print("Cancelled.")
            return None
        # Map values (subject, session, run) from session config to experiment config.
        config = map_session_config_to_experiment_config(
            session_config=session_config,
            experiment_config=config,
        )
        if mode == "data_collection_text":
            experiment_text(config=config)
        elif mode == "demo_text":
            text_demo(config=config)
        else:
            raise AssertionError
    elif mode == "eeg_to_image":
        experiment_eeg_to_image_v1(config=config)
    elif mode == "eeg_to_image_autoregressive":
        experiment_eeg_to_image_v1_autoregressive(config=config)
    elif mode == "eeg_to_image_live_demo":
        run_live_demo(cache=input_file_path)  # Pickle file path

    return None


if __name__ == "__main__":
    main()
